from arcade.glui.opengl import fs_emu_blending, fs_emu_texturing
from OpenGL import GL as gl

from arcade.glui.render import Render
from arcade.notification import Notification

START_X = 20
START_Y = 1080 - 60 - 20


# noinspection PyAttributeOutsideInit
class NotificationRender(object):
    font = None
    y = START_Y

    @classmethod
    def init(cls):
        pass
        # import pygame.font
        # font_path = resources.resource_filename(
        #     "LiberationSans-Bold.ttf")
        # cls.font = pygame.font.Font(
        #     font_path, int(0.022 * Render.get().display_height))
        # #Notification("New device connected:\nLogitech Gamepad F310",
        # duration=60)
        # #Notification("Test 2", duration=10)
        # #Notification("Test 3")

    @classmethod
    def update(cls):
        pass
        # if Notification.new:
        #     Notification.new = False
        #     Render.get().dirty = True
        #     # post wake-up event
        #     pygame.event.post(pygame.event.Event(pygame.NUMEVENTS - 1))

    @classmethod
    def render(cls):
        notifications = Notification.all()
        if len(notifications) == 0:
            return
        Render.get().hd_perspective()
        cls.y = START_Y
        for i, notification in enumerate(notifications):
            cls.render_one(notification)
            Render.get().dirty = True
            # showing max four notifications
            if i > 4:
                break

    @classmethod
    def render_one(cls, notification):
        notification.show()
        w = 500
        h = 100
        x = START_X
        y = cls.y
        z = 0.9

        y -= h

        gl.glDisable(gl.GL_DEPTH_TEST)
        fs_emu_texturing(False)
        fs_emu_blending(True)
        gl.glBegin(gl.GL_QUADS)
        gl.glColor(1.0, 1.0, 0.5, 0.8)
        gl.glVertex3f(x, y, z)
        gl.glVertex3f(x + w, y, z)
        gl.glVertex3f(x + w, y + h, z)
        gl.glVertex3f(x, y + h, z)
        gl.glEnd()

        lines = notification.text.split("\n")
        y += h - 23
        for line in lines:
            tw, th = Render.get().measure_text(line, cls.font)
            y -= th
            Render.get().text(
                line, cls.font, x + 23, y, color=(0.2, 0.2, 0.0, 1.0)
            )
            y -= 2
        # cls.y += h + 15
        cls.y -= h + 20
        gl.glEnable(gl.GL_DEPTH_TEST)



from __future__ import absolute_import
from __future__ import print_function

from .block.CommentBlock import CommentBlock
from .block.EntryBlock import EntryBlock
from .FileName import FileName
from .MetaInfo import MetaInfo
from .ProtectFlags import ProtectFlags
from .TimeStamp import TimeStamp
from .FSError import *
from .FSString import FSString
import amitools.util.ByteSize as ByteSize

class ADFSNode:
  def __init__(self, volume, parent):
    self.volume = volume
    self.blkdev = volume.blkdev
    self.parent = parent
    self.block_bytes = self.blkdev.block_bytes
    self.block = None
    self.name = None
    self.valid = False
    self.meta_info = None
  
  def __str__(self):
    return "%s:'%s'(@%d)" % (self.__class__.__name__, self.get_node_path_name(), self.block.blk_num)
  
  def set_block(self, block):  
    self.block = block
    self.name = FileName(FSString(self.block.name), is_intl=self.volume.is_intl,is_longname=self.volume.is_longname)
    self.valid = True
    self.create_meta_info()
    
  def create_meta_info(self):
    comment = self.block.comment
    if self.block.comment_block_id != 0:
        comment_block = CommentBlock(self.blkdev, self.block.comment_block_id)
        comment_block.read()
        comment = comment_block.comment
    self.meta_info = MetaInfo(self.block.protect, self.block.mod_ts, FSString(comment))

  def get_file_name(self):
    return self.name

  def delete(self, wipe=False, all=False, update_ts=True):
    if all:
      self.delete_children(wipe, all, update_ts)
    self.parent._delete(self, wipe, update_ts)

  def delete_children(self, wipe, all, update_ts):
    pass

  def get_meta_info(self):
    return self.meta_info

  def change_meta_info(self, meta_info):
    dirty = False

    # dircache?
    rebuild_dircache = False
    if self.volume.is_dircache and self.parent != None:
      record = self.parent.get_dircache_record(self.name.get_ami_str_name())
      if record == None:
        raise FSError(INTERNAL_ERROR, node=self)
    else:
      record = None
        
    # alter protect flags
    protect = meta_info.get_protect()
    if protect != None and hasattr(self.block, 'protect'):
      self.block.protect = protect
      self.meta_info.set_protect(protect)
      dirty = True
      if record != None:
        record.protect = protect

    # alter mod time
    mod_ts = meta_info.get_mod_ts()
    if mod_ts != None:
      self.block.mod_ts = mod_ts
      self.meta_info.set_mod_ts(mod_ts)
      dirty = True
      if record != None:
        record.mod_ts = mod_ts
    
    # alter comment
    comment = meta_info.get_comment()
    if comment != None and hasattr(self.block, "comment"):
      if EntryBlock.needs_extra_comment_block(self.name.get_ami_str_name(), comment.get_ami_str()):
        if self.block.comment_block_id == 0:
          # Allocate and initialize extra block for comment
          blks = self.volume.bitmap.alloc_n(1)
          if blks is not None:
            cblk = CommentBlock(self.blkdev, blks[0])
            cblk.create(self.block.blk_num)
            self.block.comment_block_id = cblk.blk_num
            self.volume.bitmap.write_only_bits()
          else:
            raise FSError(NO_FREE_BLOCKS, node=self)
        else:
          cblk = CommentBlock(self.blkdev, self.block.comment_block_id)
          cblk.read()
        cblk.comment = comment.get_ami_str()
        cblk.write()
      else:
        self.block.comment = comment.get_ami_str()
        if self.block.comment_block_id != 0:
          self.volume.bitmap.dealloc_n([self.block.comment_block_id])
          self.block.comment_block_id = 0
          self.volume.bitmap.write_only_bits()

      self.meta_info.set_comment(comment)
      dirty = True
      if record != None:
        rebuild_dircache = len(record.comment) < comment 
        record.comment = comment.get_ami_str()
    
    # really need update?
    if dirty:
      self.block.write()
      # dirache update
      if record != None:
        self.parent.update_dircache_record(record, rebuild_dircache)        
      
  def change_comment(self, comment):
    self.change_meta_info(MetaInfo(comment=comment))
    
  def change_protect(self, protect):
    self.change_meta_info(MetaInfo(protect=protect))
    
  def change_protect_by_string(self, pr_str):
    p = ProtectFlags()
    p.parse(pr_str)
    self.change_protect(p.mask)
    
  def change_mod_ts(self, mod_ts):
    self.change_meta_info(MetaInfo(mod_ts=mod_ts))
    
  def change_mod_ts_by_string(self, tm_str):
    t = TimeStamp()
    t.parse(tm_str)
    self.change_meta_info(MetaInfo(mod_ts=t))

  def get_list_str(self, indent=0, all=False, detail=False):
    istr = u'  ' * indent
    if detail:
      extra = self.get_detail_str()
    else:
      extra = self.meta_info.get_str_line()
    return u'%-40s       %8s  %s' % (istr + self.name.get_unicode_name(), self.get_size_str(), extra)
    
  def list(self, indent=0, all=False, detail=False, encoding="UTF-8"):
    print(self.get_list_str(indent=indent, all=all, detail=detail).encode(encoding))

  def get_size_str(self):
    # re-implemented in derived classes!
    return ""
    
  def get_blocks(self, with_data=False):
    # re-implemented in derived classes!
    return 0

  def get_file_data(self):
    return None

  def dump_blocks(self, with_data=False):
    blks = self.get_blocks(with_data)
    for b in blks:
      b.dump()

  def get_node_path(self, with_vol=False):
    if self.parent != None:
      if not with_vol and self.parent.parent == None:
        r = []
      else:
        r = self.parent.get_node_path()
    else:
      if not with_vol:
        return []
      r = []
    r.append(self.name.get_unicode_name())
    return r

  def get_node_path_name(self, with_vol=False):
    r = self.get_node_path()
    return FSString(u"/".join(r))

  def get_detail_str(self):
    return ""
    
  def get_block_usage(self, all=False, first=True):
    return (0,0)

  def get_file_bytes(self, all=False, first=True):
    return (0,0)

  def is_file(self):
    return False
  
  def is_dir(self):
    return False

  def get_info(self, all=False):
    # block usage: data + fs blocks
    (data,fs) = self.get_block_usage(all=all)
    total = data + fs
    bb = self.blkdev.block_bytes
    btotal = total * bb
    bdata = data * bb
    bfs = fs * bb
    prc_data = 10000 * data / total
    prc_fs = 10000 - prc_data
    res = []
    res.append("sum:    %10d  %s  %12d" % (total, ByteSize.to_byte_size_str(btotal), btotal))
    res.append("data:   %10d  %s  %12d  %5.2f%%" % (data, ByteSize.to_byte_size_str(bdata), bdata, prc_data / 100.0))
    res.append("fs:     %10d  %s  %12d  %5.2f%%" % (fs, ByteSize.to_byte_size_str(bfs), bfs, prc_fs / 100.0))
    return res



# Copyright (c) 2008 Duncan Fordyce
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

""" contains helper functions for common irc commands """

import random


def join(client, channel):
    # Dummy function - replaced later
    pass


def part(client, channel):
    # Dummy function - replaced later
    pass


def nick(client, nick):
    # Dummy function - replaced later
    pass


def msg(cli, user, msg):
    for line in msg.split('\n'):
        cli.send("PRIVMSG", user, ":%s" % line)


def msgrandom(cli, choices, dest, user=None):
    o = "%s: " % user if user else ""
    o += random.choice(choices)
    msg(cli, dest, o)


def _makeMsgRandomFunc(choices):
    def func(cli, dest, user=None):
        msgrandom(cli, choices, dest, user)
    return func

msgYes = _makeMsgRandomFunc(['yes', 'alright', 'ok'])
msgOK = _makeMsgRandomFunc(['ok', 'done'])
msgNo = _makeMsgRandomFunc(['no', 'no-way'])


def ns(cli, *args):
    msg(cli, "NickServ", " ".join(args))


def cs(cli, *args):
    msg(cli, "ChanServ", " ".join(args))


def identify(cli, passwd, authuser="NickServ"):
    msg(cli, authuser, "IDENTIFY %s" % passwd)


def quit(cli, msg='gone'):
    cli.send("QUIT :%s" % msg)
    cli._end = 1


def user(cli, username, realname=None):
    cli.send("USER", username, cli.host, cli.host, realname or username)


_simple = (
    'join',
    'part',
    'nick',
    'notice',
)


def _addsimple():
    import sys
    def simplecmd(cmd_name):
        def f(cli, *args):
            cli.send(cmd_name, *args)
        return f
    m = sys.modules[__name__]
    for t in _simple:
        setattr(m, t, simplecmd(t.upper()))


_addsimple()


# noinspection PyPep8Naming
def _addNumerics():
    import sys
    from oyoyo import ircevents

    def numericcmd(cmd_num, cmd_name):
        def f(cli, *args):
            cli.send(cmd_num, *args)
        return f

    m = sys.modules[__name__]
    for num, name in ircevents.numeric_events.items():
        setattr(m, name, numericcmd(num, name))


_addNumerics()



import fsui
from fsgs.context import fsgs
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig


class CustomOptionsPage(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        label = fsui.MultiLineLabel(
            self,
            gettext(
                "You can write key = value pairs here to set FS-UAE options "
                "not currently supported by the user interface. This is only a "
                "temporary feature until the GUI supports all options "
                "directly."
            ),
            760,
        )
        self.layout.add(label, fill=True, margin_bottom=10)
        label = fsui.MultiLineLabel(
            self,
            gettext(
                "The options specified here will apply to this configuration "
                "only."
            ),
            760,
        )
        self.layout.add(label, fill=True, margin_bottom=10)

        hor_layout = fsui.HorizontalLayout()
        self.layout.add(hor_layout, fill=True, expand=True)

        # hor_layout.add_spacer(20)
        self.text_area = fsui.TextArea(self, font_family="monospace")
        self.text_area.set_min_width(760)
        self.text_area.set_min_height(400)
        self.text_area.set_text(initial_text())
        hor_layout.add(self.text_area, fill=True, expand=True)
        # hor_layout.add_spacer(20)

        # self.layout.add_spacer(20)

        self.get_window().add_close_listener(self.on_close_window)

    def on_close_window(self):
        self.update_config()

    def on_close_button(self):
        self.end_modal(0)

    def update_config(self):
        text = self.text_area.get_text()
        update_config = {}
        # First mark all unknown config values as cleared
        for key in list(fsgs.config.values.keys()):
            if key not in LauncherConfig.default_config:
                update_config[key] = ""
        # Then we overwrite with specific values
        for line in text.split("\n"):
            line = line.strip()
            parts = line.split("=", 1)
            if len(parts) == 2:
                key = parts[0].strip()
                # if key in Config.no_custom_config:
                #     continue
                value = parts[1].strip()
                update_config[key] = value
        # Finally, set everything at once
        LauncherConfig.set_multiple(update_config.items())


def initial_text():
    text = []
    keys = fsgs.config.values.keys()
    for key in sorted(keys):
        # FIXME: Move to LauncherConfig as a method, maybe use
        # is_custom_option.
        if key in LauncherConfig.no_custom_config:
            continue
        if key.startswith("__implicit_"):
            continue

        value = fsgs.config.values[key]
        if not value:
            continue
        text.append("{0} = {1}\n".format(key, value))
    return "".join(text)



from __future__ import absolute_import, division, print_function

import os
import re
import unittest

from fsgs.util.gamenameutil import GameNameUtil


class DiskNameUtil(object):
    @classmethod
    def find_disks(
        cls, path, base_names=None, script=None, file_list=None, black_list={}
    ):
        disks = [path]
        print("")
        print("FIRST DISK", path)
        print("")
        # disks = []
        if script:
            use_name = script.rename_file(os.path.basename(path))
        else:
            use_name = os.path.basename(path)
        number = GameNameUtil.find_number(use_name)
        disk_count = GameNameUtil.find_disk_count(use_name)
        first_without_number = GameNameUtil.strip_number(use_name)
        first_without_flags = GameNameUtil.strip_flags(first_without_number)
        if number is None:
            # only this disk
            return disks
        base_name = GameNameUtil.find_base_name(path)
        dir = os.path.dirname(path)
        candidates = {}
        items = []
        if base_names:
            items = base_names[base_name]
        else:
            if file_list:
                all_items = file_list
            else:
                all_items = os.listdir(dir)
            for item in all_items:
                b = GameNameUtil.find_base_name(item)
                if base_name == b:
                    items.append(item)
        # print("items is", items)
        for item in items:
            if script:
                use_name = script.rename_file(item)
            else:
                use_name = item
            # print(item)
            if GameNameUtil.is_bad_dump(use_name):
                continue
            if GameNameUtil.find_disk_count(use_name) != disk_count:
                continue
            n = GameNameUtil.find_number(use_name)
            if n == 0:
                n = 99
            if n:
                candidates.setdefault(n, []).append(
                    (use_name, os.path.join(dir, item))
                )
        # print(candidates)
        for n, items in candidates.items():
            if n == 1:
                # already added first floppy
                continue
            print("")
            print("Candidates:")
            for item in items:
                print(item[0])
            print("")
            print("FIND FLOPPY NUMBER", n)
            print("")
            matches = []
            for use_name, p in items:
                without_number = GameNameUtil.strip_number(use_name)
                # print(without_number, "vs", first_without_number)
                if without_number == first_without_number:
                    # print("perfect match:", p)
                    # disks.append(p)
                    matches.append((-1000, p))
                    continue
                    # break
                without_flags = GameNameUtil.strip_flags(without_number)
                if without_flags == first_without_flags:
                    flags_1 = extract_flags(first_without_number)
                    cr_flag_1 = ""
                    for flag in flags_1:
                        if flag.startswith("cr"):
                            cr_flag_1 = flag

                    flags_n = extract_flags(without_number)
                    cr_flag_n = ""
                    for flag in flags_n:
                        if flag.startswith("cr"):
                            cr_flag_n = flag

                    score = 0

                    # if cr_flag_n and cr_flag_1 != cr_flag_n:
                    if cr_flag_1 and cr_flag_n and cr_flag_1 != cr_flag_n:
                        # not same cracker
                        score += 10000

                    if "[o" in without_number:
                        score += 20000

                    flag_set_1 = set(flags_1)
                    flag_set_n = set(flags_n)

                    extra_flags_in_1 = flag_set_1.difference(flag_set_n)
                    extra_flags_in_n = flag_set_n.difference(flag_set_1)

                    # score += len(extra_flags_in_1)
                    print(extra_flags_in_1)
                    for flag in extra_flags_in_1:
                        score += 10 - flags_1.index(flag)
                    score += len(extra_flags_in_n) * 100
                    print(score, p)
                    matches.append((score, p))

                    # if without_flags == without_number:
                    #     # there were no flags on this floppy
                    #     matches.append((1, p))
                    # else:
                    #     # slightly worse score since there were flags
                    #     # not matching floppy 1
                    #     matches.append((100, p))

            if len(matches) == 0:
                # raise Exception("no candidates for floppy " + num)
                # print("WARNING: choosing partial matching floppy", p,
                #         "for floppy number", n)
                # disks.append(p)
                print("Did not find good match for floppy", n)
                print("candidates:")
                for item in items:
                    print("  ", item)
                raise Exception(
                    "Did not find good match for floppy {0}".format(n)
                )
            matches.sort()

            print("")
            print("Matches:")
            for match in matches:
                print(match[0], match[1])
            print("")
            score, p = matches[0]
            # if score == 2:
            disks.append(p)
            # TOSEC (x of y) disk number labelling format
            if " of {0})".format(n) in path:
                # found the correct number of disks
                break
        print("")
        print("Result:")
        for disk in disks:
            print(disk)
        print("")
        return disks


flags_pattern = re.compile(r"\[([^\]]+)\]")


def extract_flags(name):
    matches = flags_pattern.findall(name)
    return [x.strip() for x in matches]


class TestGameDiskUtil(unittest.TestCase):
    def set_disks(self, s):
        self.all_disks = [x.strip() for x in s.split("\n") if x.strip()]

    def find_disks(self, name):
        return DiskNameUtil.find_disks(name, file_list=self.all_disks)

    def test_batman_the_movie(self):
        self.set_disks(BATMAN_THE_MOVIE)
        disks = self.find_disks(
            "Batman - The Movie (1989)(Ocean)(PAL)"
            "(Disk 1 of 2)[cr Black Monks].adf"
        )
        print(disks)

    def test_batman_the_movie_cr_qtx(self):
        self.set_disks(BATMAN_THE_MOVIE)
        disks = self.find_disks(
            "Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr QTX].adf"
        )
        self.assertEquals(
            disks,
            [
                "Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr QTX].adf",
                "Batman - The Movie (1989)(Ocean)(PAL)(Disk 2 of 2)[cr QTX].adf",
            ],
        )


BATMAN_THE_MOVIE = """
Batman - The Movie (1989)(Data East)(US)[cr].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr Band][t +4 Band].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr Band][t +4 Band][a].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr Black Monks].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr Dragons].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr QTX].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr QTX][a2].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr QTX][a3].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr QTX][a].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr QTX][f NTSC QTX].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr QTX][f NTSC].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr QTX][h Phantasie - Bandidos].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr QTX][h Phantasie].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr QTX][h TFS].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr QTX][t +4 Band].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 1 of 2)[cr QTX][t +4 Band][a].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 2 of 2)[cr Band][t +4 Band].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 2 of 2)[cr Band][t +4 Band][a].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 2 of 2)[cr Dragons].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 2 of 2)[cr QTX].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 2 of 2)[cr QTX][a bootblock].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 2 of 2)[cr QTX][a2 bootblock].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 2 of 2)[cr QTX][a3].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 2 of 2)[cr QTX][a4].adf
Batman - The Movie (1989)(Ocean)(PAL)(Disk 2 of 2)[cr QTX][a5].adf
Batman - The Movie (1989)(Ocean)(PAL)[cr][a][one disk version].adf
Batman - The Movie (1989)(Ocean)(PAL)[cr][one disk version].adf
Batman - The Movie (1991)(Hit Squad, The)(PAL).adf
"""


if __name__ == "__main__":
    unittest.main()



import fsui
from fsgs.application import ApplicationMixin


class Application(ApplicationMixin, fsui.Application):
    instance = None
    name = None

    @classmethod
    def init(cls, name):
        cls.name = name

    def __init__(self):
        super().__init__(Application.name)
        self.set_icon(fsui.Icon("fs-uae-arcade", "pkg:launcher"))



import atexit
import subprocess
import sys
import time


class Server:
    def __init__(self, port, players, password):
        self.port = port
        self.players = players
        self.password = password
        self.process = None

    def start(self):
        args = [sys.executable] + sys.argv[:]
        args.append("--server")
        args.append("--port={0}".format(self.port))
        args.append("--players={0}".format(self.players))
        if self.password:
            args.append("--password={0}".format(self.password))
        print("start gamepad config, args =", args)
        self.process = subprocess.Popen(args, close_fds=True)
        # a quick hack to give the server time to start, should instead
        # read stdout or something else
        time.sleep(2.0)
        atexit.register(self.kill)

    def kill(self):
        print("Server.kill")
        if self.process is not None:
            self.process.kill()
        self.process = None



"""
This module contains information about files present on common Workbench
disks. The data is used when creating a Workbench environment on demand.
Some files are removed because they differ between Cloanto WB floppies and
original WB floppies.
"""

import hashlib
import os
import sys
import traceback

from fsgs.amiga.adf import ADFFile

wb_133_startup_sequence = """\
c:SetPatch >NIL: ;patch system functions
Addbuffers df0: 10
cd c:
echo "A500/A2000 Workbench disk.  Release 1.3.3 version 34.34*N"
Sys:System/FastMemFirst ; move C00000 memory to last in list
BindDrivers
SetClock load ;load system time from real time clock (A1000 owners should
              ;replace the SetClock load with Date
FF >NIL: -0 ;speed up Text
resident CLI L:Shell-Seg SYSTEM pure add; activate Shell
resident c:Execute pure
mount newcon:
;
failat 11
run execute s:StartupII ;This lets resident be used for rest of script
wait >NIL: 5 mins ;wait for StartupII to complete (will signal when done)
;
SYS:System/SetMap usa1 ;Activate the ()/* on keypad
path ram: c: sys:utilities sys:system s: sys:prefs add ;set path for Workbench
LoadWB delay  ;wait for inhibit to end before continuing
endcli >NIL:
"""


wb_204_startup_sequence = """\
c:setpatch >NIL:
c:version >NIL:
addbuffers >NIL: df0: 15
Failat 21

Resident >NIL: C:Execute PURE ADD

makedir ram:T ram:Clipboards ram:env ram:env/sys
copy >NIL: ENVARC: ram:env all quiet noreq

assign ENV: ram:env
assign T: ram:t ;set up T: directory for scripts
assign CLIPS: ram:clipboards
assign REXX: s:

if exists sys:Monitors
    join >NIL: sys:monitors/~(#?.info) as t:mon-start
    execute t:mon-start
    delete >NIL: t:mon-start
endif

BindDrivers

setenv Workbench $Workbench
setenv Kickstart $Kickstart

IPrefs

echo "Amiga Release 2.  Kickstart $Kickstart, Workbench $Workbench"

conclip

mount speak:
mount aux:
mount pipe:

path ram: c: sys:utilities sys:rexxc sys:system s: sys:prefs sys:wbstartup add
if exists sys:tools
    path sys:tools add
    if exists sys:tools/commodities
        path sys:tools/commodities add
    endif
endif

; If this is the initial boot (i.e. keyboard env variable is not set)
; then execute PickMap which will query for a keymap and set the
; keyboard env variable.
;
; if keyboard env variable is set, set the keymap
if ${sys/keyboard} NOT EQ "*${sys/keyboard}"
    setmap ${sys/keyboard}
else
    PickMap sys: initial
endif

if exists s:user-startup
    execute s:user-startup
endif

LoadWB

endcli >NIL:
"""


wb_300_startup_sequence = """\
; $VER: startup-sequence 39.9 (9.8.92)

C:SetPatch QUIET
C:Version >NIL:
C:AddBuffers >NIL: DF0: 15
FailAt 21

C:MakeDir RAM:T RAM:Clipboards RAM:ENV RAM:ENV/Sys
C:Copy >NIL: ENVARC: RAM:ENV ALL NOREQ

Resident >NIL: C:Assign PURE
Resident >NIL: C:Execute PURE

Assign >NIL: ENV: RAM:ENV
Assign >NIL: T: RAM:T
Assign >NIL: CLIPS: RAM:Clipboards
Assign >NIL: REXX: S:
Assign >NIL: PRINTERS: DEVS:Printers
Assign >NIL: KEYMAPS: DEVS:Keymaps
Assign >NIL: LOCALE: SYS:Locale
Assign >NIL: LIBS: SYS:Classes ADD
Assign >NIL: HELP: LOCALE:Help DEFER

IF NOT EXISTS SYS:Fonts
  Assign FONTS:
EndIF

BindDrivers
C:Mount >NIL: DEVS:DOSDrivers/~(#?.info)

IF EXISTS DEVS:Monitors
  IF EXISTS DEVS:Monitors/VGAOnly
    DEVS:Monitors/VGAOnly
  EndIF

  C:List >NIL: DEVS:Monitors/~(#?.info|VGAOnly) TO T:M LFORMAT "DEVS:Monitors/%s"
  Execute T:M
  C:Delete >NIL: T:M
EndIF

SetEnv Workbench $Workbench
SetEnv Kickstart $Kickstart
UnSet Workbench
UnSet Kickstart

C:IPrefs

C:ConClip

Path >NIL: RAM: C: SYS:Utilities SYS:Rexxc SYS:System S: SYS:Prefs SYS:WBStartup SYS:Tools SYS:Tools/Commodities

IF EXISTS S:User-Startup
  Execute S:User-Startup
EndIF

Resident Execute REMOVE
Resident Assign REMOVE

C:LoadWB
EndCLI >NIL:
"""

# noinspection SpellCheckingInspection
wb_133_files = {
    "c/": "",
    "c/AddBuffers": "b4935ca3f55fbf53db41a61d32cee48d93d1c431",
    "c/Ask": "20156be306fadfe45b8fde83bbc90258f371a00c",
    "c/Assign": "52579fe27010466442e928cb2a62cecc20f9f184",
    "c/Avail": "9d3d8ff67cfc4d7e220db53268830b0a5b5f0a2b",
    "c/Binddrivers": "0f7eaf913a9a8d50af3573a695a8f4af3b8f0ad6",
    "c/Break": "7c1cb7e382d36feb70b773b6cfd030ff853057e1",
    "c/CD": "9f4c181e7d5e3c0e103f462ba1420570499ca922",
    "c/ChangeTaskPri": "f63a5652aacd198b6ee25a2aefa3518a53f14ab6",
    "c/Copy": "ef236641b73d0b849f6ceb4a218469ea8f1d06f2",
    "c/Date": "287859f1c0fc0f1772c9e9c4f767bd523515b194",
    "c/Delete": "d6304c0e704c94073ad67edd19725a4742697ef6",
    "c/Dir": "c4c80a92749862fe8fe877dff0d7ab88a824026d",
    "c/DiskChange": "37665418b614fb1d768ea85c32a165cfaec146c1",
    # "c/DiskDoctor": "c801fbdcc4b04ff492a67cfeaee0036fce70bb85",
    "c/Echo": "b43fc80f2fb23af7c3ee14c8e1b5eecfc3c587ee",
    # "c/Ed": "dad70eb4a65febb74283c2a189d6bb95b22b0886",
    "c/Edit": "d83ee8a2e1f9d1697875366dafc329f8e8e4d7e9",
    "c/Else": "8f64d3d3fc7c84e02c4211e68ca74b2fa528df7b",
    "c/EndCLI": "24a5167194970a0188c00b234a12d2fba4308478",
    "c/EndIf": "7f15af4934c472489f278268667c2788c2a7021b",
    "c/EndSkip": "7f15af4934c472489f278268667c2788c2a7021b",
    "c/Eval": "5af8c120e9cf1399740dab3eca367af242d015cc",
    "c/Execute": "29b7093beae26deab1bd66a98167bfdb96aa7cf7",
    "c/Failat": "49918ba167529cef1320452d7442eac73712e0e7",
    "c/Fault": "e7f20bbe32843bca9c206cab8be50ae1d77b680f",
    # "c/FF": "dc9b9750c2a38546f7be2dbe983ae258f2496b98",
    "c/FileNote": "36d14d7468eb136b1179f04fa3c218832638e960",
    "c/GetEnv": "be841e3d1794a11b9af1b299ff1d8c9be85d8338",
    "c/IconX": "442df04019ae6ff426c3464f8be8c523ad292830",
    "c/If": "e0f0c4880f33ab782d903663535812f777362983",
    "c/Info": "a22a7a4aa07bad698b725253bd796c6fbdca9eb8",
    "c/Install": "1792a60e0564098429838d8db592b6297769d6f9",
    "c/Join": "0e8d276bb1f70a52c608cd8c8315ef8e84cceaef",
    "c/Lab": "7f15af4934c472489f278268667c2788c2a7021b",
    "c/List": "0d5c1cddab2af3e41bfe485e4a8cb76dc99c4bc4",
    "c/LoadWB": "13f61f0d60028256cc790f4cef3bcca83fb20259",
    "c/Lock": "f8a6d2e97b637c1ac1be97eaf77e3f12c79154ac",
    "c/Makedir": "d01c425e8af0f4f4084a6b2ee13780240db4f027",
    "c/Mount": "4f18b1c83c0f7a59eaa11c08e7a82cb7941740b3",
    "c/NewCLI": "9298e77f0adaa7a789d6e5f9f3035f24e60364d6",
    "c/NewShell": "7b827b5fa5456c9e1d722f2611437864be1bfdec",
    "c/Path": "770f775d73fdd8df8c1890be2ba2f9fbd388c06b",
    "c/Prompt": "5bb558e3f2ed62d5f676e84bc23f70ed913b4412",
    "c/Protect": "fec1b09d6a808b0e101ee2ed923a7f0e60597caa",
    "c/Quit": "e8f9244859b2e55e5c10304b1ca35e5801265c23",
    "c/Relabel": "496bc49e1488e8276e2074063b839269b34d1889",
    "c/RemRAD": "bca2dbf0ea2a8885fd976c04d421b884796d070a",
    "c/Rename": "b94baf0954756f75272576c7b34cdd6859d896d3",
    "c/Resident": "70c21c5a6add12526b29690f8bde0327f91b6f78",
    "c/Run": "caf92a05af573c54a1c9044e51ac6aeaab534e23",
    "c/Search": "0359aa4849bb1e7f7fa37862a4ab2f34b6d3b6fb",
    # "c/SetClock": "95d98f533207f53c1d90ddc57f2815db56d9bba7",
    "c/SetDate": "cca8245571bd7ffc4ddd46e87441b2600bf8be74",
    "c/SetEnv": "ac8df5ae6adfc7d68e5a3f514b935bd5293b09b7",
    "c/SetPatch": "3b36ee49e10626594545c564957fb4882ac2bfc3",
    "c/Skip": "15c513c48b568dc77593c13bf593fc6f8e0bcdb8",
    "c/Sort": "39f4a8a41ec5ab3cb030be579b49b6a3a7046f5a",
    "c/Stack": "d48dc30db5f9d6c7c04259444ba74b7c2f3dffd0",
    "c/Status": "08fffe1acc5dec49345e188eaba1f2fa74cd7424",
    "c/Type": "afb53409c675bac58048e8e76d403d61e91c1fc9",
    "c/Version": "bfa3c28746dc9a56df4569e687d0c7f6bb7e5b80",
    "c/Wait": "c893b36bd2376c003e3d2b2f5791d757c5fa33f9",
    "c/Which": "d1c3a0b2d9a0bcc3db3580182b75961981e9d38d",
    "c/Why": "84d318064b8374399bad8139c5f8f6046aa0b6e0",
    "devs/": "",
    "devs/clipboard.device": "d1708898f40bdb4f525e615e833e7f109331ac56",
    "devs/clipboards/": "",
    "devs/keymaps/": "",
    "devs/keymaps/usa1": "39d79e8a775260c13df0f657ee20e5f3e742de9e",
    "devs/MountList": "b85692e12b2fc7eaa02e4d211b043b5cd6272200",
    "devs/narrator.device": "d51631238dc07925b89e57ebe9d2e412400a0384",
    "devs/parallel.device": "861b780af555f52088adf5106bb62c05444dba82",
    "devs/printer.device": "5f06ec8cd7c32c40d78af95c8953ec527647a861",
    "devs/printers/": "",
    "devs/printers/generic": "aaaab28025bef46e5c01dde3d01d2f03fc63df2b",
    "devs/ramdrive.device": "ca60b73bc779d95cdcf23ec5e92b44dc21b1b472",
    "devs/serial.device": "55df2ec0a1d9dc6a03c9335737e23a38b2d73213",
    "devs/system-configuration": "f4e0ca14cbfb69c78c491d4cb4253a9a8ba9a78b",
    "Empty/": "",
    "Expansion/": "",
    "fonts/": "",
    "fonts/diamond.font": "cfe171bc8ab48615c6ac8fdfc5cd9da098f35e66",
    "fonts/diamond/": "",
    "fonts/diamond/12": "e6634c30e031dae19428423952f0abbe75d84559",
    "fonts/diamond/20": "75cb8014065834d1be7261a265bf0939e95d483a",
    "fonts/emerald.font": "98d935aa47ca406384bb6c592c63549f2d43d604",
    "fonts/emerald/": "",
    "fonts/emerald/17": "359d8c7da2e5ff4c157eed528c8c41a5ea09530a",
    "fonts/emerald/20": "de377bd498532ff4d3aa92e60fae8a029e7fc608",
    "fonts/garnet.font": "57e054da85337f25d660d2eb983dcf924a44ca4d",
    "fonts/garnet/": "",
    "fonts/garnet/16": "e99e9fde239b557ba181d91261f84abfda6f8bcb",
    "fonts/garnet/9": "5705e5553c85dc2e6573e4420d704e9fca5efe25",
    "fonts/opal.font": "dc66cd2d713f027b22539c503c5a5647e1f29fbf",
    "fonts/opal/": "",
    "fonts/opal/12": "ebb7b174ba5f215172921ec7e8cb3774eb21c90b",
    "fonts/opal/9": "0cea9720d43e2280b6fa8583fcefa364b9540446",
    "fonts/ruby.font": "314e2c9ca75cb75cadca1038d19a5ab92f2ed656",
    "fonts/ruby/": "",
    "fonts/ruby/12": "c5b1fd7d3834ebd2a26fc23c42f174451ddea1c7",
    "fonts/ruby/15": "58e3f44c65d75c9bff0a9992536c084c2572dbc9",
    "fonts/ruby/8": "a09b9ce5bd1c4a68773e6abfccc5218d0ddd066d",
    "fonts/sapphire.font": "18f992c2fef46fd408b272f07f9d5598c5e61bcb",
    "fonts/sapphire/": "",
    "fonts/sapphire/14": "fb0da752206e8b29e94f6e3d68492c2f794c54d5",
    "fonts/sapphire/19": "250e989ac01401b10773ad11cf9e45ad83437fe6",
    "fonts/topaz.font": "66b2ef88f256216f340a2130920c3dafdc5b3038",
    "fonts/topaz/": "",
    "fonts/topaz/11": "be925d2d0a61962ff1de9deb742747a356fc2b87",
    "l/": "",
    "l/Aux-Handler": "2e621414176ce93959e066cacc86f16c70bb7f80",
    "l/Disk-Validator": "30ede410472723c95338ed2c870916ad275e9706",
    "l/FastFileSystem": "cc14a3ce4805d4e3a0397fe6997ff5428d90ecdf",
    "l/Newcon-Handler": "c530e2ce63f2c6203526345fb2d434f2d4a9f49d",
    "l/Pipe-Handler": "4e73c34bcf73ab5b590c7b016dc1176673322077",
    "l/Port-Handler": "d7e4ddab8cbd94f751a6ffd4bc93cfe07cb4cae2",
    "l/Ram-Handler": "c0b036019fbffc46b417a72dd296f8d33b19a308",
    "l/Shell-Seg": "4bbe3ccb55fab49c83ff7ab52b2af277e6d92429",
    "l/Speak-Handler": "1b3b3dcd0b46cc6dbf98467e319d0d396e7df22f",
    "libs/": "",
    "libs/diskfont.library": "f1a4d45d97f2df1cff6b83a4392c34cc7ac18bbc",
    "libs/icon.library": "4c9b95e6f786f707aa51b2483a9f21ac5e66d7c9",
    "libs/info.library": "aa46c186d762dc68d2e799a5d13cad6745b55d10",
    "libs/mathieeedoubbas.library": "cf565b2b7a9f6e8aae030e32490346589f90b780",
    "libs/mathieeedoubtrans.library": "29f906dc50594aca3a91cfb69bd3f332fdb1671f",
    "libs/mathtrans.library": "65bae4aac804fb4db79a427f12b6ef1e16d22882",
    "libs/translator.library": "6ddc4462f261317a61f3c371091021b243df562c",
    "libs/version.library": "8f062571cac294bf5556b5bdc9d48da2c2f21fae",
    "Prefs/": "",
    "Prefs/CopyPrefs": "0af96045315193d4ec4f6c40d5ea6c3765d3fb04",
    "Prefs/Preferences": "e8b98bbb4cc3a04cf31855d4dcf6c292d02a48b4",
    "s/": "",
    "s/CLI-Startup": "c9abb720ab20b5976e18c45dd6eee46187b367f9",
    "s/DPAT": "bd0ee7066b43ef7df98db2801a75ab4f79f13098",
    "s/PCD": "d6ff57db9f9cd193d99aa64581ca964494b5e430",
    # "s/Shell-Startup": "25ee071ea6069e7fb10ac23bef1bbb24426b1bf1",
    "s/SPAT": "208cd7d832fb8a0d81d45fc3db7bc31c12727b3f",
    "s/Startup-Sequence.HD": "0d3caa59e5f867a6f9f8c68d318a16da1bf60fd6",
    # "s/StartupII": "dae0c0aeb988f27c5a9a85dcdf502998570b3b45",
    "Shell": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
    "System/": "",
    "System/CLI": "a6e7eb1585e4ebd6f4ed606eb18d8f23af184bea",
    "System/Diskcopy": "009cfa0952888c059da450b46d731d7f1779ae63",
    "System/FastMemFirst": "5fd51854d5b109fcdbf52809497e642036aa4722",
    "System/FixFonts": "3fc2dfbc62326678c3daf9ee860bac6330666999",
    # "System/Format": "972bbf474d471bec395d19f6d737908542055d70",
    "System/InitPrinter": "6d35a5215846acef053211dfcd6563054b51da10",
    "System/MergeMem": "04c018c68442187d292c0a57e1686bc8269f2ac4",
    "System/NoFastMem": "87e8c332e69a5446535923157cb6678ccb992614",
    "System/SetMap": "6d6c0ca1163c9d921b193d0e3c04722800c8c88d",
    "t/": "",
    "Trashcan/": "",
    "Utilities/": "",
    "Utilities/Calculator": "66d0b0fcb4795517c6d30cb96662afce68d81585",
    "Utilities/Clock": "ca791b988dd647432dd36dcc519bc5ce7d2e0bd2",
    "Utilities/ClockPtr": "47c191dd0edeccbf80cb5402002101bcc791e171",
    "Utilities/CMD": "7e0917b6fbf4152fd9167af0eee2118b35d48fe7",
    "Utilities/GraphicDump": "b82aa41aa1d602a1e1452d757797dd22b97dc75a",
    "Utilities/InstallPrinter": "05f57e5ffc717c7a4eb80bb3917be6006ebdc0b0",
    "Utilities/More": "2eb6ab87d1ee9bf28bea826e2481dd3c4f6c1447",
    "Utilities/Notepad": "ca9e25cd122dc01a28b66eb8f3ebc3395504e6b7",
    "Utilities/PrintFiles": "66d70dae78836d1aa198b9fd46492f204be0459f",
    "Utilities/Say": "3f845bf4d06c680c960847115449adb4819cd8b7",
}

# noinspection SpellCheckingInspection
wb_204_files = {
    "C/": "",
    "C/AddBuffers": "6ed1a3f24a1f605aa7234410faebbfbb3fda3ee6",
    "C/Assign": "b5b7edb67f578019d46af425a16458ec0cdb1c2e",
    "C/Avail": "5e62c202fe3cb447a19c39efb43c935eed139a66",
    "C/BindDrivers": "7534362e5def2eb32819be52f242d28ebdc05a26",
    "C/Break": "97ca2404e17fc78e3d988131bb17183c5c3aa15a",
    "C/ChangeTaskPri": "66533bb2cee36475b03cd73ae8ab198889e81be3",
    "C/ConClip": "96dd0671b40743c18f9ec2b58b2ccb731092e29f",
    "C/Copy": "628a17ad2b883565a1312d105a65764076c42013",
    "C/CPU": "40bcedc1af7677096cb1e5f5e41e147cf4c952fc",
    "C/Date": "0d190c0ce0b99cd653a8f982964eed110ac540eb",
    "C/Delete": "0597fd916dd28850ef68bb904937cad0d347cf5a",
    "C/Dir": "184eef562670fd6b22b3025e385f992469fa50d4",
    "C/DiskChange": "2cae12bfbb1cc98348e7afabf0f48b09cadbaf5e",
    "C/DiskDoctor": "c801fbdcc4b04ff492a67cfeaee0036fce70bb85",
    "C/Ed": "077483aacde3b40eeeb7b2b08cff3c18d6174e1c",
    "C/Edit": "14a1b22e4d72f85cd64b49cfdc8380b294f2da15",
    "C/Eval": "fae9792514d07b6fbb5b04d778492941dee5c448",
    "C/Execute": "ef1e6f6db0f3a513e0dc98add537e1db87b1856a",
    "C/Filenote": "6c263e034a3aca677467d410e410ba808a5f2b8d",
    "C/IconX": "9a2088f98175b3341f8bf4831599d54832430c35",
    "C/Info": "d12eea32aef864fcb4717ee8447c01d5900c94f1",
    "C/Install": "b5e1a2ea287c29390d38e47ccb3749f7c449bd4e",
    "C/IPrefs": "f3eb8607aaac99043020f0a9a1156dec4f941a11",
    "C/Join": "da8d787eadd16e682216c808cbbd1427acde7280",
    "C/List": "74a07847e45f8ae61daa029dd21775f438013a08",
    "C/LoadWB": "c705d1579c04505819a6bc3c6c29ae8d2610da66",
    "C/Lock": "a671aabf698196baa6a237601d07fb49b7f64f60",
    "C/MagTape": "8d14c533f6fb73286a642bb208151c36da3af2c9",
    "C/MakeDir": "3e00b210538357ff2ca9f0bc16471f06fb994282",
    "C/MakeLink": "da8472a553ece5484d7ba1290884ad5d59db5770",
    "C/Mount": "b2f1e59e977cec1c83166977cd4350f5388c3c9f",
    "C/Protect": "e95073c0169dd7c7ca39b9842b339a1288ca3016",
    "C/Relabel": "056cb6a12033f5638b1353be1007106fd9267a6e",
    "C/RemRAD": "7e5d53c865debaa9d752f96ef818320248332ef6",
    "C/Rename": "d1cb2ab35c8fa0dc912b98dbcf002c2c3b220988",
    "C/Search": "89c1dd049ca2fff58269fd44af2705abdedff112",
    "C/SetClock": "0c166dd6f8baa45eda766f3b41a0f799f8e7f388",
    "C/SetDate": "d7a78803ade6e09dbbba051ab6d4fbcb923527eb",
    "C/SetFont": "8b5bd918009242703f5f8ae9237199bd35ef0efe",
    "C/SetPatch": "309db203c5c88d2867d9d6f75e5ef67d2c558a95",
    "C/Sort": "d886f81bdc74f4942acb52680e56d8bf50a0be8a",
    "C/Status": "71e09cfc79d8ede0721ba4a01e1cd43a6318a99f",
    "C/Type": "00917f88e78770f6b28562bede2cf64718325553",
    "C/Version": "86dcf19836869c7319a1b89083a0529d1ee0d049",
    "C/Wait": "a0d503a136dad028ec51e33cfd57d130c1b4b807",
    "C/Which": "22952cce8446a1405b1b22d057ec19614e38c366",
    "Devs/": "",
    "Devs/clipboard.device": "83f3490b6480a0515b37e6f926ab48daf9cad61e",
    "Devs/Keymaps/": "",
    "Devs/MountList": "b32b62007cbfd4e5412f003f3fd83a183845b447",
    # "Devs/narrator.device": "c5a35d605c39b9c59542d71883238a02e92d726d",
    "Devs/parallel.device": "88c8c2bae625caf7b2bdbea9688a763f035bc199",
    "Devs/printer.device": "b13d1ec922deafa5a479fc1c591f911ff95559eb",
    "Devs/Printers/": "",
    "Devs/Printers/generic": "11473c5271b970e36c70b408aac7c67fff7d97b6",
    "Devs/serial.device": "d814d6ce8efb4a87966dae581eddf4dca528094a",
    "Devs/system-configuration": "4be3d7e8395a5827085f57cb6cdbf5e88fa78506",
    "Expansion/": "",
    "Fonts/": "",
    "L/": "",
    "L/aux-handler": "5a20c44cdcba0793fd292e3f951832ad4670f65e",
    "L/port-handler": "d7e4ddab8cbd94f751a6ffd4bc93cfe07cb4cae2",
    "L/queue-handler": "a458c6935c90d8a9422600c84875237e0558f89b",
    # "L/speak-handler": "d577708e1a0be7566885824349102025f1a250c4",
    "Libs/": "",
    "Libs/asl.library": "8d1fd81d7c128c397443f0ceb696dab3fecc5828",
    "Libs/commodities.library": "eaa02d69480d8df876f3932bf13f7c6e6ebc6c78",
    "Libs/diskfont.library": "97022049498794fe8f6135b53ed01d0688903499",
    "Libs/iffparse.library": "32eb189c8c003e8bdb1c836fa0ffd12a5d2c9f17",
    "Libs/mathieeedoubbas.library": "ce7888086d9749d738e1141e7b7e200f5eb780a9",
    "Libs/mathieeedoubtrans.library": "6d6e29a25f7bc87d26a56d437441d3a2c789b126",
    "Libs/mathieeesingtrans.library": "b9b164b6a7bff61ffd703999c93f756ea447452f",
    "Libs/mathtrans.library": "92d5888b3d2d3bb83c66cc6e486d303821c518c9",
    "Libs/rexxsupport.library": "7ae7acdd99a89d00b706060f818ad340869827a2",
    "Libs/rexxsyslib.library": "b74995c09a0d6350498579b8ff406477ae5b9566",
    # "Libs/translator.library": "e99c035faf5184a3397c89a3df1a7606c8417be6",
    "Libs/version.library": "ac699c82157ccc204aaf1eb78e3c53c4a8f13bf5",
    "Monitors/": "",
    "Prefs/": "",
    "Prefs/Env-Archive/": "",
    "Prefs/Env-Archive/sys/": "",
    "Prefs/Env-Archive/sys/wbconfig.prefs": "9658314fdb2a32286dba8830cf701469ac0089d3",
    "Prefs/Font": "c8d0f7bd565fb151ff0738e9df9f063cd43cf244",
    "Prefs/IControl": "c3d984f4ca7213194c8bfa5cb20ef8487c107e7a",
    "Prefs/Input": "37b19e8dd679c1d901c6620d0635cd9a56b92ab9",
    "Prefs/Overscan": "d15843c3147bff57f3dd4ffc2303d1609ca0eb12",
    "Prefs/Palette": "3d8712e5b337e8a02c5e786975b5724bd9716f8e",
    "Prefs/Pointer": "79c013f9d2065b7c2f824e530c05e65b182c00de",
    "Prefs/Presets/": "",
    "Prefs/Printer": "89a003a1b26c5a0b9eb2fdf6ddd53cb5368445cb",
    "Prefs/PrinterGfx": "84d71c99acac85cf65cf073e736b55871e286d2c",
    "Prefs/ScreenMode": "640adc69430a72f08a42761b4d3740840d95a834",
    "Prefs/Serial": "8b7e24eb43396dc0fc5a341a2425213fb7d77ac3",
    "Prefs/Time": "bb07a47dbb7453fbeb61ecae5ba9459f33b4217c",
    "Prefs/WBPattern": "42b809a0d4f02ea5e7f720df3e7a5e97827a1689",
    "Rexxc/": "",
    "Rexxc/HI": "d3934c81b5a7832f0cdd64650f7c74eacc608a0a",
    "Rexxc/RX": "527adec943412976b2d40c4556ca48e0a5b0abee",
    "Rexxc/RXC": "b42afa06c53df3221683a2d42b2b2ce4f38a4525",
    "Rexxc/RXLIB": "04b10ce3d05feb2fdbd722e5a275a0c844e7d86e",
    "Rexxc/RXSET": "702b4613236e8faba2270744229171dc46a0d5ed",
    "Rexxc/TCC": "473c7136e636b9505332118cd69a3ba29db0dbbe",
    "Rexxc/TCO": "f8b901a6ec8c6844a7205f5ceec5aad6e6261531",
    "Rexxc/TE": "2f89c6cd66e559b6f3cf5a0783b14fb18236931d",
    "Rexxc/TS": "45a5aa31dd4d42099d41948ef976599fa221938d",
    "Rexxc/WaitForPort": "f3fea65223cb028f479d1831237385812675a065",
    "S/": "",
    "S/BRUtab": "36cad75345610b07e447edb5000a368ce14de0df",
    "S/DPat": "4c8aac6a4989201c01d36247a9622288beb6d291",
    "S/Ed-startup": "298221ea95f4bfa2cc9b0cec4d8643f708a16abc",
    "S/HDBackup.config": "5284e3e2897148a373da769df16530cd1fff45c6",
    "S/PCD": "d6ff57db9f9cd193d99aa64581ca964494b5e430",
    "S/PickMap": "0b9ca76ba793031e11dfd14cf26b1257f9b24d89",
    "S/Shell-startup": "3d50527c6c17256f50d04d212d97d0fb27bf862f",
    "S/SPat": "870ac2f50dbb2a75989c5418a2012df81455bea8",
    "S/Startup-sequence.HD": "d66c044a25bc4235ec1d2479f368aca733a4d77d",
    "System/": "",
    "System/AddMonitor": "9ef54a726a70298608654a8be35388adbbd4a5eb",
    "System/BindMonitor": "5bafd3b20970c9cb05e9619f28091d56ad1dc2c8",
    "System/CLI": "1d22d9100a26d12d375fe1fd938f46b5c09017f1",
    "System/DiskCopy": "c91ae2a853c386c6c4c2c5dbed394bc69074f552",
    "System/FixFonts": "5f093b674d951599af5b5b2f82f3007cfdfa2902",
    "System/Format": "e8f87083aefbaf68f03947accca10278b95ad9d5",
    "System/NoFastMem": "abb2cff5c39351d111bce07c22585d58219b31a1",
    "System/RexxMast": "6f64f8bc51aec68344fae4151697976f59946b8e",
    "System/Setmap": "859abaaab6de713c2a46043f5f5421d54b79d88f",
    "t/": "",
    "Trashcan/": "",
    "Utilities/": "",
    "Utilities/Clock": "079dadb61258bbb07112a4109b58552fe61f5741",
    "Utilities/Display": "5c21232ceea79568d915d2d9d0a96e80e01fd4cb",
    "Utilities/Exchange": "b227f5823b2b07f92670f2628426645fbfa4ec40",
    "Utilities/More": "39d44f82ad645b9b1cd2931125c5b340613f7e29",
    "Utilities/Say": "79614ec17080d87d580f6cd1e20e74f53c1b5174",
    "WBStartup/": "",
}

# noinspection SpellCheckingInspection
wb_300_files = {
    "C/": "",
    "C/AddBuffers": "6ed1a3f24a1f605aa7234410faebbfbb3fda3ee6",
    "C/AddDataTypes": "e8bbd0b50fb5de6fa2382e845e1d67011a1ddd2a",
    "C/Assign": "b5b7edb67f578019d46af425a16458ec0cdb1c2e",
    "C/Avail": "5e62c202fe3cb447a19c39efb43c935eed139a66",
    "C/BindDrivers": "be270c04b4591f10da960231e488929d67e58135",
    "C/Break": "97ca2404e17fc78e3d988131bb17183c5c3aa15a",
    "C/ChangeTaskPri": "66533bb2cee36475b03cd73ae8ab198889e81be3",
    "C/ConClip": "956fd130af649ac441d4b68ea9b01e29d4861357",
    "C/Copy": "0c5f9470fbcb36fdbcdf970fd1b22aa627328a24",
    "C/CPU": "0191712ede6348b8e1803f5807a941a85c12bd0a",
    "C/Date": "0d190c0ce0b99cd653a8f982964eed110ac540eb",
    "C/Delete": "0597fd916dd28850ef68bb904937cad0d347cf5a",
    "C/Dir": "184eef562670fd6b22b3025e385f992469fa50d4",
    "C/DiskChange": "2cae12bfbb1cc98348e7afabf0f48b09cadbaf5e",
    "C/Ed": "077483aacde3b40eeeb7b2b08cff3c18d6174e1c",
    "C/Edit": "14a1b22e4d72f85cd64b49cfdc8380b294f2da15",
    "C/Eval": "fae9792514d07b6fbb5b04d778492941dee5c448",
    "C/Execute": "ef1e6f6db0f3a513e0dc98add537e1db87b1856a",
    "C/Filenote": "6c263e034a3aca677467d410e410ba808a5f2b8d",
    "C/IconX": "09bd52d33b700538e9bed493f245a3ffce944ad3",
    "C/Info": "3f5f37e405ca929cf0b4c6ac6abbf7f8d4f19892",
    "C/Install": "b4bf5e4ecda66c6a66e1780d1143f632e573ed3b",
    "C/IPrefs": "5b5b70ec7b06a6fa142d01372a4e0fce1bff5461",
    "C/Join": "da8d787eadd16e682216c808cbbd1427acde7280",
    "C/List": "86c0a28542939ea5e0d2075f9db337d93b139fb7",
    "C/LoadWB": "e663a715c6ad69a7d3882b4772cbdc9de835e791",
    "C/Lock": "a671aabf698196baa6a237601d07fb49b7f64f60",
    # "C/MagTape": "8d14c533f6fb73286a642bb208151c36da3af2c9",
    "C/MakeDir": "3e00b210538357ff2ca9f0bc16471f06fb994282",
    "C/MakeLink": "da8472a553ece5484d7ba1290884ad5d59db5770",
    "C/Mount": "c691df2ceb10c6dccd4f735dea0f07c199e09074",
    "C/Protect": "e95073c0169dd7c7ca39b9842b339a1288ca3016",
    "C/Relabel": "056cb6a12033f5638b1353be1007106fd9267a6e",
    "C/RemRAD": "7e5d53c865debaa9d752f96ef818320248332ef6",
    "C/Rename": "d1cb2ab35c8fa0dc912b98dbcf002c2c3b220988",
    "C/RequestChoice": "5198856e4ebc9c88e5799627aeceaa42866f6525",
    "C/RequestFile": "46a45ca9906b52a215161c2141193fe740f41d71",
    "C/Search": "89c1dd049ca2fff58269fd44af2705abdedff112",
    "C/SetClock": "0c166dd6f8baa45eda766f3b41a0f799f8e7f388",
    "C/SetDate": "d7a78803ade6e09dbbba051ab6d4fbcb923527eb",
    "C/SetFont": "3b2bb7bb70d84a230a4614dae34c064b7cfc315d",
    "C/SetKeyboard": "d32a3aed9f10fdcbb4335296be50f70f2f15438e",
    "C/SetPatch": "4d4aae988310b07726329e436b2250c0f769ddff",
    "C/Sort": "d886f81bdc74f4942acb52680e56d8bf50a0be8a",
    "C/Status": "71e09cfc79d8ede0721ba4a01e1cd43a6318a99f",
    "C/Type": "00917f88e78770f6b28562bede2cf64718325553",
    "C/Version": "5a698d6494fa50fd3faa7e532def17b6c561d217",
    "C/Wait": "a0d503a136dad028ec51e33cfd57d130c1b4b807",
    "C/Which": "22952cce8446a1405b1b22d057ec19614e38c366",
    "Classes/": "",
    "Classes/DataTypes/": "",
    "Classes/DataTypes/8svx.datatype": "dae1f0ff6171479b27c21d3e32f7d4e383f969ab",
    "Classes/DataTypes/amigaguide.datatype": "d53c3e6abd119a3ca7c15805642a1e6ec96d374c",
    "Classes/DataTypes/ascii.datatype": "6f54362a5ce75221a62e0f1fd3a5d23fabc0f2bd",
    "Classes/DataTypes/ilbm.datatype": "c8251ed31ad79aa71cf6d3c134815c8f38c70226",
    "Classes/DataTypes/picture.datatype": "8028e6782c4069c45ffc4b51cef493f8339a1be9",
    "Classes/DataTypes/sound.datatype": "e4fbdbc5e55ddc4635dcd2e76e5623217c4a41a2",
    "Classes/DataTypes/text.datatype": "e6b80a188123cb397f7a79a1b580206bb369f2e2",
    "Classes/Gadgets/": "",
    "Classes/Gadgets/colorwheel.gadget": "7bbfccb8fd5d68b2ae8aac226eff877e8ebc9734",
    "Classes/Gadgets/gradientslider.gadget": "a54f39e9098f2c58e892b9183b555fe0a8574191",
    "Devs/": "",
    "Devs/clipboard.device": "c8d85bd384ea5033d5474283e83c02ad6cdaf32b",
    "Devs/DataTypes/": "",
    "Devs/DataTypes/8SVX": "b9116bb7654e12f7d281d32ca333a0f283e15532",
    "Devs/DataTypes/AmigaGuide": "e89b0bf0d75585b8f27e40d711164ab9e2a0c222",
    "Devs/DataTypes/FTXT": "cee60b9550ace3d44ba1d214ed350b05587c3919",
    "Devs/DataTypes/ILBM": "1cf3598ca839c554e6a7acc333ddd8327f8ac5d3",
    "Devs/DOSDrivers/": "",
    "Devs/DOSDrivers/PIPE": "624dd68e9a2794bf5bd60e58c1b3679a1dfd273c",
    "Devs/Keymaps/": "",
    # "Devs/mfm.device": "63b150010e420f96304375badca126812daa2255",
    "Devs/Monitors/": "",
    "Devs/parallel.device": "88c8c2bae625caf7b2bdbea9688a763f035bc199",
    "Devs/postscript_init.ps": "3c0bc3408a4ac929936f647637a04f45d227cca0",
    "Devs/printer.device": "aaf3970e54bedaf84cd69268fc6cf3730c395ba1",
    "Devs/Printers/": "",
    "Devs/Printers/Generic": "11473c5271b970e36c70b408aac7c67fff7d97b6",
    "Devs/serial.device": "d814d6ce8efb4a87966dae581eddf4dca528094a",
    "Devs/system-configuration": "4be3d7e8395a5827085f57cb6cdbf5e88fa78506",
    "Expansion/": "",
    "L/": "",
    # "L/aux-handler": "5a20c44cdcba0793fd292e3f951832ad4670f65e",
    # "L/CrossDOSFileSystem": "9f05f997b3aa0c3a431bea96cc4bbc153ba48814",
    # "L/FileSystem_Trans/": "",
    # "L/FileSystem_Trans/DANSK.crossdos":
    # "3db3dbfdb9f6d874368f12152944b00099fd9943",
    # "L/FileSystem_Trans/INTL.crossdos":
    # "6d803e82923564dd2f12b68836b2356282edfd93",
    "L/port-handler": "d7e4ddab8cbd94f751a6ffd4bc93cfe07cb4cae2",
    "L/queue-handler": "a458c6935c90d8a9422600c84875237e0558f89b",
    "Libs/": "",
    "Libs/amigaguide.library": "42b1ea9b94f12f4ad5275bb29a8a49fb977f1910",
    "Libs/asl.library": "ae59765242d7d1fc022d819a980652463e72688a",
    "Libs/bullet.library": "efffa71e955805d543842cbd0c67e75b591fe567",
    "Libs/commodities.library": "e52e0a7e2fb653a6895fb1cf15d0b37aed18f909",
    "Libs/datatypes.library": "1ea18282089ae14620c1fc71e82592d52217ff2a",
    "Libs/diskfont.library": "15d9765f1b66c4a3068d11186ea5bf41a8c8ad3b",
    "Libs/iffparse.library": "946ef60b4ba9f63593dbadfbc447546d7ab9725c",
    "Libs/locale.library": "a12b91f662a1c527ff6411b43555bc7eaeb370b4",
    "Libs/mathieeedoubbas.library": "08d8508cdcb77ad421a6dd6f80721b727c09d96b",
    "Libs/mathieeedoubtrans.library": "6d6e29a25f7bc87d26a56d437441d3a2c789b126",
    "Libs/mathieeesingtrans.library": "b9b164b6a7bff61ffd703999c93f756ea447452f",
    "Libs/mathtrans.library": "92d5888b3d2d3bb83c66cc6e486d303821c518c9",
    "Libs/rexxsupport.library": "7ae7acdd99a89d00b706060f818ad340869827a2",
    "Libs/rexxsyslib.library": "b74995c09a0d6350498579b8ff406477ae5b9566",
    "Libs/version.library": "9af4fa21ce77ca97b4604ffb095a9b2488295c85",
    "Prefs/": "",
    "Prefs/Env-Archive/": "",
    "Prefs/Env-Archive/Sys/": "",
    "Prefs/Env-Archive/Sys/wbconfig.prefs": "9658314fdb2a32286dba8830cf701469ac0089d3",
    "Prefs/Presets/": "",
    "Rexxc/": "",
    "Rexxc/HI": "d3934c81b5a7832f0cdd64650f7c74eacc608a0a",
    "Rexxc/RX": "527adec943412976b2d40c4556ca48e0a5b0abee",
    "Rexxc/RXC": "b42afa06c53df3221683a2d42b2b2ce4f38a4525",
    "Rexxc/RXLIB": "04b10ce3d05feb2fdbd722e5a275a0c844e7d86e",
    "Rexxc/RXSET": "702b4613236e8faba2270744229171dc46a0d5ed",
    "Rexxc/TCC": "473c7136e636b9505332118cd69a3ba29db0dbbe",
    "Rexxc/TCO": "f8b901a6ec8c6844a7205f5ceec5aad6e6261531",
    "Rexxc/TE": "2f89c6cd66e559b6f3cf5a0783b14fb18236931d",
    "Rexxc/TS": "45a5aa31dd4d42099d41948ef976599fa221938d",
    "Rexxc/WaitForPort": "f3fea65223cb028f479d1831237385812675a065",
    "S/": "",
    # "S/DPat": "227c6ed4ae33850577a0b03e22b7dfd966b421c4",
    "S/Ed-startup": "2a4e2dd940726199aea3c5cab73dcc527b624fb8",
    # "S/PCD": "02d4b5292a9ea3b56d68ca7bba3d1df5055b3a25",
    "S/Shell-Startup": "d039a940c89ec6c30a1b3568bfd6afe626a28b6d",
    # "S/SPat": "4f63e17cb9ca68b5dc5eb37fd02508f4c5ae0761",
    "System/": "",
    "System/CLI": "9f25538fd0a6b134dd7a094350731cdc1e9234e3",
    "System/DiskCopy": "902aa22b272961911a279bba4aca1fd7ba9c537b",
    "System/FixFonts": "86757c1120ce2137e3c72024fa9de898de8a7b24",
    "System/Format": "76918338279f5c1a2b86405f39d9e9c8e9ac0981",
    "System/NoFastMem": "6295d9937d3eb3370af14bf510c9ad02f687bd4b",
    "System/RexxMast": "6f64f8bc51aec68344fae4151697976f59946b8e",
    "T/": "",
    "Utilities/": "",
    "Utilities/Clock": "fda1ad2f1bfb730d9c0dbda4b1141935b67f65ae",
    "Utilities/More": "daced6e93fce3bf91a659ea161e343b87233cc54",
    "Utilities/MultiView": "c8e7b9f35907e168a03f5b46c9890b4094077750",
    "WBStartup/": "",
}

# noinspection SpellCheckingInspection
wb_133_floppies = [
    # Workbench v1.3.3 rev 34.34 (1990)(Commodore)(A500-A2000)
    # (Disk 1 of 2)(Workbench).adf
    "86cb4e007f9fdcf6c5eda6adba3f60f188063875",
    # Workbench v1.3.3 rev 34.34 (1990)(Commodore)(A500-A2000)(Dk)
    # (Disk 1 of 2)(Workbench)[m].adf
    "8d8314faa3b5fbc472c11d5fc669358522c1d00b",
    # amiga-os-134-workbench.adf
    "42c5af6554212e9d381f7535c3951ee284e127b2",
]

# noinspection SpellCheckingInspection
wb_204_floppies = [
    # Workbench v2.04 rev 37.67 (1991)(Commodore)
    # (Disk 1 of 4)(Workbench).adf
    "8d5c0310a86f14fb3e6a1da001ceb50b9a592c51",
    # Workbench v2.04 rev 37.67 (1991)(Commodore)
    # (Disk 1 of 4)(Workbench)[m].adf
    "d9d8bc1964d9159b3669fadedcd140ead197b0b7",
    # Workbench v2.04 rev 37.67 (1991)(Commodore)
    # (Disk 1 of 4)(Workbench)[m4].adf
    "21a4363e236011f0173c393207a2225a0a3002b0",
    # Workbench v2.04 rev 37.67 (1991)(Commodore)
    # (Disk 1 of 4)(Workbench)[m2].adf
    "898c0c372c476d8410890388b81ca642cc0b381d",
    # Workbench v2.04 rev 37.67 (1991)(Commodore)
    # (Disk 1 of 4)(Workbench)[m3].adf
    "5913fa0fb6cfa74ae9c80870f6ab8a4289036788",
    # amiga-os-204-workbench.adf
    "898c0c372c476d8410890388b81ca642cc0b381d",
]

# noinspection SpellCheckingInspection
wb_300_floppies = [
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 2 of 6)(Workbench)[!].adf
    "e663c92a9c88fa38d02bbb299bea8ce70c56b417",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 2 of 6)(Workbench)[m2].adf
    "cf2f24cf5f5065479476a38ec8f1016b1f746884",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 2 of 6)(Workbench)[m5].adf
    "4f4770caae5950eca4a2720e0424df052ced6a32",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 2 of 6)(Workbench)[a].adf
    "9496daa66e6b2f4ddde4fa2580bb3983a25e3cd2",
    # Workbench v3.0 rev 39.29 (1992)(Commodore)(A1200-A4000)(M10)
    # (Disk 2 of 6)(Workbench)[m3].adf
    "0e7f30223af254df0e2b91ea409f35c56d6164a6",
    # amiga-os-300-workbench.adf
    "4f4770caae5950eca4a2720e0424df052ced6a32",
]


def update_files():
    base_dir = os.path.expanduser("~/Documents/FS-UAE/Floppies/Workbench")
    # floppies = {}
    floppy_file_sets = {}
    files = {}
    for floppy in os.listdir(base_dir):
        if not floppy.lower().endswith(".adf"):
            continue
        path = os.path.join(base_dir, floppy)
        with open(path, "rb") as f:
            floppy_sha1 = hashlib.sha1(f.read()).hexdigest()
        sha1sums = set()
        try:
            adf = ADFFile(path)
        except Exception:
            print("error parsing", path)
            traceback.print_exc()
            continue
        names = adf.namelist()
        for name in names:
            if name.endswith("/"):
                continue
            data = adf.open(name, "r").read()
            sha1 = hashlib.sha1(data).hexdigest()
            sha1sums.add(sha1)
            # floppies[floppy.lower()] = os.path.join(base_dir, floppy)
            files.setdefault(sha1, set()).add((floppy_sha1, name))
        floppy_file_sets[(floppy, floppy_sha1)] = sha1sums

    interesting_files = set()
    for file_map in [wb_204_files, wb_300_files]:
        for name, sha1 in file_map.items():
            if sha1:
                interesting_files.add(sha1)

    # print("")
    # print("workbench_file_map = {")
    # for sha1 in sorted(interesting_files):
    #     print("    \"{0}\": [".format(sha1))
    #     #print("        \"{0}\",".format(floppy_sha1))
    #     for floppy_sha1, name in sorted(files[sha1]):
    #         #print("        \"{0}\",".format(floppy_sha1))
    #         print("        (\"{0}\", \"{1}\"),".format(floppy_sha1, name))
    #     print("    ],")
    # print("}")
    # print("")

    for name, file_map in [
        ("wb_133_floppies", wb_133_files),
        ("wb_204_floppies", wb_204_files),
        ("wb_300_floppies", wb_300_files),
    ]:
        print("\n" + name + " = [")
        file_set = set([x for x in file_map.values() if x])
        for floppy_data, floppy_file_set in floppy_file_sets.items():
            floppy_name, floppy_sha1 = floppy_data
            if floppy_file_set.issuperset(file_set):
                print("    # {0}".format(floppy_name))
                print('    "{0}",'.format(floppy_sha1))
        print("]")


def main():
    if sys.argv[1] == "--update-files":
        update_files()
        return

    base_dir = os.path.expanduser("~/Documents/FS-UAE/Floppies/Workbench")
    floppies = {}
    for floppy in os.listdir(base_dir):
        if not floppy.lower().endswith(".adf"):
            continue
        floppies[floppy.lower()] = os.path.join(base_dir, floppy)

    floppy = floppies[sys.argv[1].lower()]
    adf = ADFFile(floppy)
    names = adf.namelist()
    startup_sequence = ""
    for name in names:
        if name.lower() == "s/startup-sequence":
            startup_sequence = adf.open(name, "r").read()
            continue
        if name.endswith(".info"):
            continue
        if name.endswith("/"):
            sha1 = ""
        else:
            data = adf.open(name, "r").read()
            sha1 = hashlib.sha1(data).hexdigest()
        if sha1:
            if len(name) > (80 - 4 - 2 - 1 - 2 - 2 - 40):
                print('    "{0}":\n    "{1}",'.format(name, sha1))
            else:
                print('    "{0}": "{1}",'.format(name, sha1))
        else:
            print('    "{0}": "",'.format(name))
    print("")
    print(startup_sequence)


if __name__ == "__main__":
    main()



import os

import fsboot
import fsui
from fsui.qt import QFont


class Font(fsui.Font):
    def __init__(self, name, size):
        qfont = QFont(name)
        qfont.setPixelSize(size)
        super().__init__(qfont)

    @classmethod
    def stream(cls, name):
        path = os.path.join(fsboot.base_dir(), "Workspace", "Fonts", name)
        try:
            return open(path, "rb")
        except FileNotFoundError:
            raise NotImplementedError()



import re
import subprocess
import sys
import traceback

from fsgs.amiga.fsuaedevicehelper import FSUAEDeviceHelper
from launcher.i18n import gettext
from launcher.launcher_settings import LauncherSettings
from launcher.launcher_signal import LauncherSignal
from launcher.option import Option


class Device:
    def __init__(self, id, name, type):
        self.id = id
        self.name = name
        self.type = type
        self.port = None
        self.cmp_id = Device.create_cmp_id(id)

    @staticmethod
    def create_cmp_id(id_):
        return id_.lower().replace(" ", "")

    def __repr__(self):
        return "<Device {}".format(self.id)


class DeviceManager:
    initialized = False
    devices = []
    device_ids = []
    device_names = []
    device_types = []
    device_name_count = {}

    joystick_data = {}
    sdl_names = {}

    @classmethod
    def joystick_buttons(cls, device):
        return cls.joystick_data[device][0]

    @classmethod
    def joystick_hats(cls, device):
        return cls.joystick_data[device][1]

    @classmethod
    def joystick_axes(cls, device):
        return cls.joystick_data[device][2]

    @classmethod
    def joystick_balls(cls, device):
        return cls.joystick_data[device][3]

    @classmethod
    def joystick_guid(cls, device):
        return cls.joystick_data[device][4]

    @classmethod
    def init(cls):
        if cls.initialized:
            return
        # init can be called more than once (by setting initialized to
        # false, used by refresh function, so we need to clear the lists...
        cls.devices = []
        cls.device_ids = []
        cls.device_names = []
        cls.device_types = []
        cls.device_name_count = {}

        cls.devices.append(Device("none", gettext("No Host Device"), "none"))
        # cls.devices.append(Device("mouse", _("Mouse"), "mouse"))
        cls.init_fsuae()
        for id, name, type in zip(
            cls.device_ids, cls.device_names, cls.device_types
        ):
            cls.devices.append(Device(id, name, type))
        # cls.devices.append(
        #     Device("keyboard", _("Cursor Keys and Right Ctrl/Alt"),
        #            "joystick"))
        cls.initialized = True

    @classmethod
    def refresh(cls):
        cls.initialized = False
        cls.init()

        # FIXME: REPLACE WITH EnumerateHelper!!!

        LauncherSignal.broadcast("device_list_updated")

    @classmethod
    def init_fsuae(cls):
        print("DeviceManager: finding connected joysticks")
        try:
            p = FSUAEDeviceHelper.start_with_args(
                ["--list"], stdout=subprocess.PIPE
            )
            joysticks = p.stdout.read()
            p.wait()
        except Exception:
            print("exception while listing joysticks and devices")
            traceback.print_exc()
            return
        print(repr(joysticks))
        # If the character conversion fails, replace will ensure that
        # as much as possible succeeds. The joystick in question will
        # not be pre-selectable in the launcher, but the other ones will
        # work at least.
        joysticks = joysticks.decode("UTF-8", "replace")
        joysticks = [x.strip() for x in joysticks.split("\n") if x.strip()]
        last_joystick = ""
        for line in joysticks:
            if line.startswith("#"):
                continue
            if line.startswith("Buttons:"):
                parts = line.split(" ")
                buttons = int(parts[1])
                hats = int(parts[3])
                axes = int(parts[5])
                balls = int(parts[7])
                guid = parts[9]
                cls.joystick_data[last_joystick] = (
                    buttons,
                    hats,
                    axes,
                    balls,
                    guid,
                )
                continue
            if line.startswith("SDLName:"):
                value = line.split(" ")[1]
                # Strip quotes
                cls.sdl_names[last_joystick] = value[1:-1]

            device_type, name = line.split(" ", 1)
            # if name.lower() in ["keyboard", "mouse"]:
            #     # these are automatically added
            #     continue
            name_count = cls.device_name_count.get((device_type, name), 0) + 1
            cls.device_name_count[(device_type, name)] = name_count
            if name_count > 1:
                name = name + " #" + str(name_count)
            cls.device_ids.append(name)
            name = re.sub("[ ]+", " ", name)
            cls.device_names.append(name)
            if device_type == "J:":
                device_type = "joystick"
                last_joystick = name
            elif device_type == "M:":
                device_type = "mouse"
            elif device_type == "K:":
                # works as an emulated joystick...
                # device_type = "joystick"
                device_type = "keyboard"
            cls.device_types.append(device_type)

        if "--add-dummy-joystick" in sys.argv:
            cls.device_ids.append("Dummy Joystick")
            cls.device_names.append("Dummy Joystick")
            cls.device_types.append("joystick")
            cls.joystick_data["Dummy Joystick"] = (
                1,
                0,
                2,
                0,
                "c6c1bc29b0124fe6890757bb09ef006f",
            )

    @classmethod
    def get_joystick_names(cls):
        cls.init()
        return cls.device_names[:]

    @classmethod
    def get_device_type(cls, device):
        cls.init()
        for i in range(len(cls.device_names)):
            if cls.device_names[i] == device:
                return cls.device_types[i]
        return ""

    @classmethod
    def is_joystick(cls, device):
        return cls.get_device_type(device) == "joystick"

    @classmethod
    def get_joystick_ids(cls):
        cls.init()
        return cls.device_ids[:]

    @classmethod
    def get_preferred_joysticks(cls):
        prefs = []
        if LauncherSettings.get("primary_joystick"):
            prefs.append(
                Device.create_cmp_id(LauncherSettings.get("primary_joystick"))
            )
        if LauncherSettings.get("secondary_joystick"):
            prefs.append(
                Device.create_cmp_id(
                    LauncherSettings.get("secondary_joystick")
                )
            )
        return prefs

    @classmethod
    def get_preferred_gamepads(cls):
        return cls.get_preferred_joysticks()

    @classmethod
    def get_calculated_port_mode(cls, config, port):
        value = config.get("joystick_port_{0}_mode".format(port))
        if not value:
            if port == 0:
                return "mouse"
            elif port == 1:
                if config.get("amiga_model").startswith("CD32"):
                    return "cd32 gamepad"
                else:
                    return "joystick"
            return "nothing"
        return value

    @classmethod
    def get_devices_for_ports(cls, config):
        cls.init()
        ports = [cls.devices[0] for _ in range(5)]
        for device in cls.devices:
            device.port = None
        for p in range(4):
            key = "joystick_port_{0}".format(p)
            value = config.get(key)
            for device in cls.devices:
                if device.id == value:
                    device.port = p
                    break

        # print("-")
        # for device in cls.devices:
        #     print(device.port, device.id)
        # print("-")

        def auto_fill(port, type):
            mode = config.get("joystick_port_{0}_mode".format(port))
            if not mode:
                mode = cls.get_calculated_port_mode(config, port)
            val = config.get("joystick_port_{0}".format(port))
            if val:
                # specific device chosen
                for dev in cls.devices:
                    if dev.id == val:
                        ports[port] = dev
                        break
                return
            if type == "mouse":
                # print("a", mode)
                if mode != "mouse":
                    return
                # print("b")
                for dev in cls.devices:
                    # print("c")
                    if dev.type == "mouse" and dev.port is None:
                        # print("d")
                        ports[port] = dev
                        dev.port = port
                        return
            elif type == "joystick":
                if mode == "cd32 gamepad":
                    prefs = cls.get_preferred_gamepads()
                elif mode == "joystick":
                    prefs = cls.get_preferred_joysticks()
                else:
                    return
                # try to find an available preferred device first
                for pref in prefs:
                    for dev in cls.devices:
                        if dev.cmp_id == pref and dev.port is None:
                            ports[port] = dev
                            dev.port = port
                            return
                # find first suitable device
                for dev in cls.devices:
                    if dev.type == "joystick" and dev.port is None:
                        ports[port] = dev
                        dev.port = port
                        return
                for dev in cls.devices:
                    if dev.type == "keyboard" and dev.port is None:
                        ports[port] = dev
                        dev.port = port
                        return

        for p in [0, 1, 2, 3]:
            auto_fill(p, "mouse")
        for p in [1, 0, 3, 2]:
            auto_fill(p, "joystick")
        return ports

    @classmethod
    def get_device_for_port(cls, config, port):
        return cls.get_devices_for_ports(config)[port]

    @classmethod
    def get_non_amiga_devices_for_ports(cls, config):
        platform = config.get("platform")
        cls.init()
        # Launcher has 5 input devices in total (4 + 1 virtual)
        # ports = [cls.devices[0] for _ in range(5)]
        # + 1 dummy device for index 0
        ports = [cls.devices[0] for _ in range(6)]

        for device in cls.devices:
            device.port = None
        for port_ in range(1, 4 + 1):
            key = "{}_port_{}".format(platform, port_)
            value = config.get(key)
            print(key, value)
            for device in cls.devices:
                if device.id == value:
                    device.port = port_
                    break

        print("before auto-fill")
        print(ports)

        # print("-")
        # for device in cls.devices:
        #     print(device.port, device.id)
        # print("-")

        def auto_fill(port, type):
            mode = config.get("{}_port_{}_type".format(platform, port))
            if not mode:
                # FIXME: DEFAULT
                # mode = cls.get_calculated_port_mode(config, port)
                # mode = "gamepad"
                try:
                    option = Option.get(
                        "{}_port_{}_type".format(platform, port)
                    )
                    mode = option["default"]
                except KeyError:
                    # FIXME: How to handle?
                    mode = ""

            print("mode for", port, "is", mode)
            if cls.is_mouse_device(mode):
                mode = "mouse"

            val = config.get("{}_port_{}".format(platform, port))
            if val:
                # specific device chosen
                for dev in cls.devices:
                    if dev.id == val:
                        ports[port] = dev
                        break
                return

            if type == "mouse":
                # print("a", mode)
                if mode != "mouse":
                    return
                # print("b")
                for dev in cls.devices:
                    # print("c")
                    if dev.type == "mouse" and dev.port is None:
                        # print("d")
                        ports[port] = dev
                        dev.port = port
                        return
            elif type == "joystick":
                if mode == "none":
                    return
                elif mode == "mouse":
                    return
                # elif mode == "gamepad":
                #     prefs = cls.get_preferred_gamepads()
                # elif mode == "joystick":
                #     prefs = cls.get_preferred_joysticks()
                # else:
                #     return

                prefs = cls.get_preferred_gamepads()

                # try to find an available preferred device first
                for pref in prefs:
                    for dev in cls.devices:
                        if dev.cmp_id == pref and dev.port is None:
                            ports[port] = dev
                            dev.port = port
                            return
                # find first suitable device
                for dev in cls.devices:
                    if dev.type == "joystick" and dev.port is None:
                        ports[port] = dev
                        dev.port = port
                        return
                for dev in cls.devices:
                    if dev.type == "keyboard" and dev.port is None:
                        ports[port] = dev
                        dev.port = port
                        return

        for p in [1, 2, 3, 4]:
            auto_fill(p, "mouse")
        # FIXME: Hack, circular dependency
        from fsgs.platform import Platform

        if platform == Platform.C64:
            port_order = [2, 1, 3, 4]
        else:
            port_order = [1, 2, 3, 4]
        for p in port_order:
            auto_fill(p, "joystick")
            print("auto-fill", p, "=", ports[p])
        return ports

    @classmethod
    def get_non_amiga_device_for_port(cls, config, port):
        ports = cls.get_non_amiga_devices_for_ports(config)
        print(ports, port)
        return ports[port]

    @classmethod
    def is_mouse_device(cls, type_):
        # FIXME: Should be put in platform-specific code, here for now.
        return type_ in [
            "mouse",  # Amiga, ...
            "zapper",  # NES
            "arkanoid",  # NES
        ]



import builtins
import os
import subprocess
import sys
import zipfile

import workspace.path
import workspace.ui


STATE_NORMAL = 0
STATE_SELECTED = 1


def workspace_open(path, args=None):
    if args is None:
        args = []
    print("workspace.shell.open", path)
    t = tool(path)
    print(t)
    if not t:
        t = default_tool(path)
    if t:
        args.insert(0, path)
        return open(t, args)
    else:
        name = workspace.path.basename(path)
        app_path = workspace.path.join(path, name + ".py")
        print("APP PATH", app_path)
        if workspace.path.exists(app_path):
            print("Exists!")
            return open_app(app_path, args)

    print("open", path, args)

    if path in ["C:Python", "Python"]:
        print(sys.executable)
        p = subprocess.Popen([sys.executable, "run.py", "Python"] + args)
    elif path in ["C:FS-UAE", "FS-UAE", "SYS:System/FS-UAE"]:
        # FIXME: Spawn via "FS-UAE Launcher" instead, respecting Launcher
        # settings, etc.
        from fsgs.amiga.fsuae import FSUAE

        # FIXME: Other args
        FSUAE.start_with_args([workspace.path.host(args[0])])


# FIXME:
open = workspace_open


def open_app(path, args=None):
    print("open_app", path, args)
    # FIXME: Use sys load_python_program_module
    # FIXME: Not a nice method, requires globally unique module names for
    # applications
    import importlib

    host_path = workspace.path.host(path)
    host_dir = os.path.dirname(host_path)
    if host_dir not in sys.path:
        sys.path.insert(0, host_dir)
    module = importlib.import_module(
        os.path.splitext(os.path.basename(host_path))[0]
    )
    print(module.__file__)
    if hasattr(module, "workspace_open"):
        module.workspace_open(args)
    elif hasattr(module, "shell_open"):
        module.shell_open(args)
    else:
        module.open()


def file_type_icon(file_type, state=STATE_NORMAL):
    print("get file type icon for", file_type)
    icon_name = "icon.png"
    if state:
        icon_name = "selected.png"
    try:
        return icon_from_zip(
            "SYS:Icons/FileTypes/{}.fs-info".format(file_type), icon_name
        )
    except LookupError:
        pass
    p = "SYS:Icons/FileTypes/{}.fs-info/{}".format(file_type, icon_name)
    print(p)
    if workspace.path.exists(p):
        return workspace.ui.Image(p)
    return None


def default_icon(path, state=STATE_NORMAL):
    print("get default icon for", path)
    ext = workspace.path.extension(path, prefix=True, suffix=True)
    print(ext)
    return file_type_icon(ext, state)

    # try:
    #     return icon_from_zip(
    #         "SYS:Icons/FileTypes/{}.fs-info".format(ext), icon_name)
    # except LookupError:
    #     pass
    # p = "SYS:Icons/FileTypes/{}.fs-info/{}".format(ext, icon_name)
    # print(p)
    # if workspace.path.exists(p):
    #     return workspace.ui.Image(p)
    # return None


def stream_from_zip(path, name):
    try:
        zf = zipfile.ZipFile(workspace.path.host(path))
        return zf.open(name, "r")
    except (FileNotFoundError, IsADirectoryError, KeyError):
        raise LookupError(name)


def icon_from_zip(path, icon_name):
    return workspace.ui.Image(stream_from_zip(path, icon_name))


def icon(path, state=STATE_NORMAL):
    icon_name = "icon.png"
    if state:
        icon_name = "selected.png"
    try:
        return icon_from_zip(path + ".fs-info", icon_name)
    except LookupError:
        try:
            return icon_from_zip(os.path.join(path, ".fs-info"), icon_name)
        except (LookupError, NotADirectoryError):
            pass

    # p = workspace.path.host(
    p = workspace.path.join(path + ".fs-info", icon_name)
    print(p, os.path.exists(p))
    if not workspace.path.exists(p):
        # p = workspace.path.host(
        p = workspace.path.join(path, ".fs-info", icon_name)
        print(p, os.path.exists(p))
    if workspace.path.exists(p):
        return workspace.ui.Image(p)
    # return default_icon(path, state=state)
    return None


def default_tool(path):
    ext = workspace.path.extension(path, prefix=True, suffix=True)
    if ext in ["txt", "ini", "inf", "html"]:
        return "SYS:Utilities/MultiView"
    elif ext in ["url"]:
        return "SYS:System/WebBrowser"
    elif ext in ["fs-uae"]:
        return "SYS:System/FS-UAE"
    elif ext in ["mod"]:
        return "SYS:Utilities/ModulePlayer"
    # if ext == "fs-uae":
    #     return "FS-UAE"
    # if ext == "html":
    #     return "HTMLView"
    # if ext == "py":
    #     return "Python"
    # if ext == "mod":
    #     return "ModulePlayer"
    # if ext == "url":
    #     return "WebBrowser"
    # if ext == "adf":
    #     return "FS-UAE"
    return None


def tool(path):
    p = workspace.path.host(workspace.path.join(path + ".fs-info", "tool.txt"))
    print(p, os.path.exists(p))
    if not os.path.exists(p):
        p = workspace.path.host(
            workspace.path.join(path, ".fs-info", "tool.txt")
        )
        print(p, os.path.exists(p))
    if os.path.exists(p):
        with builtins.open(p, "r", encoding="UTF-8") as f:
            text = f.read().strip()
            return text
    return None



import os

from fsgs.drivers.gamedriver import Emulator, GameDriver
from fsgs.drivers.retroarchdriver import RetroArchDriver
from fsgs.input.mapper import InputMapper
from fsgs.platform import Platform
from fsgs.platforms.loader import SimpleLoader

N64_PLATFORM_ID = "n64"
N64_PLATFORM_NAME = "Nintendo 64"
N64_CONTROLLER = {
    "type": "gamepad",
    "description": "Gamepad",
    "mapping_name": "nintendo64",
}
N64_PORTS = [
    {"description": "Input Port 1", "types": [N64_CONTROLLER]},
    {"description": "Input Port 2", "types": [N64_CONTROLLER]},
    {"description": "Input Port 3", "types": [N64_CONTROLLER]},
    {"description": "Input Port 4", "types": [N64_CONTROLLER]},
]


class Nintendo64Platform(Platform):
    # FIXME: Move to init instead
    PLATFORM_NAME = N64_PLATFORM_NAME

    def driver(self, fsgs):
        return Nintendo64RetroArchDriver(fsgs)

    def loader(self, fsgs):
        return Nintendo64Loader(fsgs)


class Nintendo64Loader(SimpleLoader):
    pass


class Nintendo64MupenDriver(GameDriver):
    PORTS = [
        {"description": "Input Port 1", "types": [N64_CONTROLLER]},
        {"description": "Input Port 2", "types": [N64_CONTROLLER]},
        {"description": "Input Port 3", "types": [N64_CONTROLLER]},
        {"description": "Input Port 4", "types": [N64_CONTROLLER]},
    ]

    def __init__(self, fsgs):
        super().__init__(fsgs)
        self.emulator = Emulator("mupen64plus")
        self.emulator.allow_system_emulator = True
        self.helper = Nintendo64Helper(self.options)

    # def force_aspect_ratio(self):
    #     return 4.0 / 3.0

    # def game_video_size(self):
    #     # FIXME
    #     if self.is_pal():
    #         size = (256, 240)
    #     else:
    #         size = (256, 224)
    #     return size

    # FIXME: Scan/index byteswapped .n64/.z64 files? (i.e. #/original
    # and #!/byte-swapped.

    def prepare(self):
        temp_dir = self.temp_dir("mupen64plus")
        self.emulator.args.extend(["--configdir", temp_dir.path])
        self.emulator.args.extend(["--datadir", temp_dir.path])
        config_file = os.path.join(temp_dir.path, "mupen64plus.cfg")
        with open(config_file, "w") as f:
            self.write_config(f)
        input_config_file = os.path.join(temp_dir.path, "InputAutoCfg.ini")
        with open(input_config_file, "wb") as f:
            pass
        rom_path = self.get_game_file()
        self.emulator.args.extend([rom_path])

    def finish(self):
        pass

    def write_config(self, f):
        f.write("[Core]\n")
        f.write("OnScreenDisplay = False\n")
        # FIXME: Save state path does not seem to work
        # -stuff are saved to default dir in AppData/Roaming/Mupen64Plus/save
        # instead...
        # FIXME: Might work now with updated mupen64plus
        f.write(
            "SaveStatePath = '{path}'\n".format(
                path=self.get_state_dir() + os.sep
            )
        )
        self.configure_audio(f)
        self.configure_input(f)
        self.configure_video(f)

    def configure_audio(self, f):
        pass

    def configure_input(self, f):
        input_mapping = {
            "DPAD_RIGHT": "DPad R",
            "DPAD_LEFT": "DPad L",
            "DPAD_DOWN": "DPad D",
            "DPAD_UP": "DPad U",
            "START": "Start",
            "Z": "Z Trig",
            "B": "B Button",
            "A": "A Button",
            "C_RIGHT": "C Button R",
            "C_LEFT": "C Button L",
            "C_DOWN": "C Button D",
            "C_UP": "C Button U",
            "R": "R Trig",
            "L": "L Trig",
            "MEMPAK": "Mempak switch",
            "RUMBLEPAK": "Rumblepak switch",
            "STICK_LEFT": ("X Axis", 0),
            "STICK_RIGHT": ("X Axis", 1),
            "STICK_UP": ("Y Axis", 0),
            "STICK_DOWN": ("Y Axis", 1),
        }
        for i, port in enumerate(self.ports):
            if port.device is None:
                continue
            mapper = Mupen64PlusInputMapper(port, input_mapping)
            config = {}
            for key, value in mapper.items():
                print("---->", key, value)
                if isinstance(key, tuple):
                    key, index = key
                else:
                    index = 0
                config.setdefault(key, {})[index] = value
            f.write("\n[Input-SDL-Control{0}]\n\n".format(i + 1))
            f.write("version = 2\n")
            # Specifies whether this controller is 'plugged in' to the
            # simulated N64.
            f.write("plugged = True\n")
            # Specifies which type of expansion pak is in the controller:
            # 1=None, 2=Mem pak, 5=Rumble pak.
            f.write("plugin = 2\n")
            # If True, then mouse buttons may be used with this controller.
            f.write("mouse = False\n")
            # Controller configuration mode:
            # 0=Fully Manual, 1=Auto with named SDL Device, 2=Fully automatic.
            f.write("mode = 0\n")
            if port.device.type == "joystick":
                f.write("device = {0}\n".format(port.device.index))
                f.write('AnalogDeadZone = "512,512"\n')
                f.write('AnalogPeak = "32767,32767"\n')
            else:
                # -2 means keyboard/mouse
                f.write("device = -2\n")

            for key, value in config.items():
                type = value[0][0]
                values = [x[1][1] for x in sorted(list(value.items()))]
                values_str = ",".join(values)
                f.write(
                    '{key} = "{type}({values})"\n'.format(
                        key=key, type=type, values=values_str
                    )
                )
                print(
                    '{key} = "{type}({values})"\n'.format(
                        key=key, type=type, values=values_str
                    )
                )

    def configure_video(self, f):
        f.write("[Video-General]\n")
        if self.use_fullscreen():
            f.write("Fullscreen = True\n")
            f.write("ScreenWidth = {0}\n".format(self.screen_size()[0]))
            f.write("ScreenHeight = {0}\n".format(self.screen_size()[1]))
        else:
            f.write("Fullscreen = False\n")
            # f.write("ScreenWidth = 320\n")
            # f.write("ScreenHeight = 240\n")

        f.write("\n[UI-Console]\n\n")
        f.write("Version = 1\n")
        video_plugin = "glide64mk2"
        # video_plugin = "glide64"
        video_plugin = "rice"
        f.write('VideoPlugin = "mupen64plus-video-{}"\n'.format(video_plugin))

        if self.configure_vsync():
            # cannot find config for vsync in rice video plugin,
            # but should work for linux/nvidia due to env. variable
            # being set
            pass

        if video_plugin == "rice":
            self.configure_video_rice(f)
        elif video_plugin == "glide64":
            self.configure_video_glide64(f)
        elif video_plugin == "glide64mk2":
            self.configure_video_glide64mk2(f)

    def configure_video_rice(self, f):
        f.write("\n[Video-Rice]\n\n")
        f.write("AccurateTextureMapping = True\n")
        f.write("ForceAlphaBlender = True\n")
        # f.write("InN64Resolution = True\n")
        # f.write("RenderToTexture = 4\n")
        f.write("ScreenUpdateSetting = 2\n")
        # f.write("TextureFilteringMethod = 0\n")
        # f.write("Mipmapping = 1\n")
        # f.write("MultiSampling = 2\n")
        # f.write("AccurateTextureMapping = False\n")
        # f.write("WinFrameMode = True\n")
        # f.write("InN64Resolution = False\n")
        # f.write("ForceTextureFilter = 0\n")
        # f.write("TextureFilteringMethod = 1\n")
        # f.write("TextureFilteringMethod = 0\n")
        # f.write("FogMethod = 1\n")
        # f.write("EnableVertexShader = True\n")
        # f.write("NormalAlphaBlender = True\n")
        # f.write("ShowFPS = True\n")
        # f.write("EnableMipmaping = False\n")

        # self.args.extend(["--set", "Video-Rice[InN64resolution]=True"])
        # self.args.extend(["--set", "Video-Rice[ForceTextureFilter]=0"])

    def configure_video_glide64(self, f):
        f.write("\n[Video-Glide64]\n\n")

    def configure_video_glide64mk2(self, f):
        f.write("\n[Video-Glide64mk2]\n\n")
        # f.write("wrpResolution = 320x200\n")
        # f.write("filtering = 0\n")


class Mupen64PlusInputMapper(InputMapper):
    def axis(self, axis, positive):
        dir_str = "+" if positive else "-"
        return "axis", "{0}{1}".format(axis, dir_str)

    def hat(self, hat, direction):
        dir_str = {
            "left": "Left",
            "right": "Right",
            "up": "Up",
            "down": "Down",
        }[direction]
        return "hat", "{0} {1}".format(hat, dir_str)

    def button(self, button):
        return "button", str(button)

    def key(self, key):
        return "key", str(key.sdl_code)


class Nintendo64RetroDriver(GameDriver):
    PORTS = N64_PORTS

    def __init__(self, fsgs):
        super().__init__(fsgs)
        self.emulator = Emulator("retroarch")
        self.emulator.allow_system_emulator = True
        self.helper = Nintendo64Helper(self.options)

    def prepare(self):
        temp_dir = self.temp_dir("mupen64plus")
        self.emulator.args.extend(["--configdir", temp_dir.path])
        self.emulator.args.extend(["--datadir", temp_dir.path])
        config_file = os.path.join(temp_dir.path, "mupen64plus.cfg")
        with open(config_file, "w") as f:
            self.write_config(f)
        input_config_file = os.path.join(temp_dir.path, "InputAutoCfg.ini")
        with open(input_config_file, "wb") as f:
            pass
        rom_path = self.get_game_file()
        self.emulator.args.extend([rom_path])

    def finish(self):
        pass


class Nintendo64RetroArchDriver(RetroArchDriver):
    PORTS = N64_PORTS

    def __init__(self, fsgc):
        super().__init__(fsgc, "mupen64plus_libretro", "RetroArch/Mupen64Plus")
        self.helper = Nintendo64Helper(self.options)

    def prepare(self):
        super().prepare()
        rom_path = self.get_game_file()
        # self.helper.fix_ines_rom(rom_path)
        self.emulator.args.extend([rom_path])

        # Workaround for Intel / MESA on Linux (FIXME: Should perhaps check if
        # Intel / MESA driver is in use first...
        # https://github.com/gonetz/GLideN64/issues/454
        self.emulator.env["MESA_GL_VERSION_OVERRIDE"] = "3.3COMPAT"
        self.emulator.env["MESA_GLSL_VERSION_OVERRIDE"] = "420"

    def retroarch_input_mapping(self, port):
        input_mapping = {
            "A": "input_player{n}_b",
            "B": "input_player{n}_y",
            "DPAD_UP": "input_player{n}_up",
            "DPAD_DOWN": "input_player{n}_down",
            "DPAD_LEFT": "input_player{n}_left",
            "DPAD_RIGHT": "input_player{n}_right",
            "STICK_UP": "input_player{n}_l_y_minus",
            "STICK_DOWN": "input_player{n}_l_y_plus",
            "STICK_LEFT": "input_player{n}_l_x_minus",
            "STICK_RIGHT": "input_player{n}_l_x_plus",
            "C_UP": "input_player{n}_r_y_plus",
            "C_DOWN": "input_player{n}_r_y_minus",
            "C_LEFT": "input_player{n}_r_x_minus",
            "C_RIGHT": "input_player{n}_r_x_plus",
            "RUMBLEPAK": "input_player{n}_select",  # ???
            "START": "input_player{n}_start",
            "L": "input_player{n}_l",
            "R": "input_player{n}_r",
            "Z": "input_player{n}_l2",
        }
        return {k: v.format(n=port + 1) for k, v in input_mapping.items()}


class Nintendo64Helper:
    def __init__(self, options):
        self.options = options



import math

from arcade.glui.bottombar import (
    render_bottom_bar,
    render_bottom_bar_transparent,
)
from arcade.glui.displaylists import DisplayLists
from arcade.glui.menu import Menu
from arcade.glui.opengl import fs_emu_blending, fs_emu_texturing
from OpenGL import GL as gl

from arcade.glui.render import Render
from arcade.glui.state import State
from arcade.glui.texture import Texture
from arcade.glui.texturemanager import TextureManager

LIGHTING = False
MIN_WRAPAROUND_ITEMS = 9


def render_menu(menu, what=None, skip_center_item=False):
    if what is None:
        what = ["ITEMS", "SHADOWS"]
    Render.get().standard_perspective()
    item_data = _render_menu(menu, what, skip_center_item)
    return item_data


def _render_menu(menu, what=None, skip_center_item=False):
    if what is None:
        what = ["ITEMS", "SHADOWS"]
    # print "render menu with", len(menu), "items"
    position = menu.get_current_position()
    # selected = menu.get_selected_index()
    assert not skip_center_item

    # min_item = 0
    # max_item = len(menu) - 1
    num_items = len(menu)
    # i_position = min(max_item, int(round(position)))
    i_position = int(round(position))

    # print position, " (", selected, ")"
    # print "selected", selected

    # glEnable(GL_DEPTH_TEST)

    def get_item_position(ip_distance):
        inverse = ip_distance < 0
        ip_distance = abs(ip_distance)
        # y = 0.02 - 0.5
        # y = 0.243
        y = 0.293
        z = -3.1
        if ip_distance < 1.0:
            x = ip_distance * 0.8
            ip_rotation = -45.0 * ip_distance
            z = z + 0.6 - 0.6 * ip_distance
        else:
            # x = 0.8 + 0.15 * (ip_distance - 1.0)
            # x = 0.8 + 0.20 * (ip_distance - 1.0)
            x = 0.8 + 0.33 * (ip_distance - 1.0)
            ip_rotation = -45.0
        if inverse:
            return [-x, y, z], -ip_rotation
        else:
            return [x, y, z], ip_rotation

    def yield_render_item_sequence():
        if not skip_center_item:
            yield i_position
        # if MenuGameTransition.value == 0: #< 0.00001:
        # for i in range(9, 0, -1):
        for i in range(1, 10):
            # if (i_position - i) >= min_item:
            yield i_position - i
            # if (i_position + i) <= max_item:
            yield i_position + i

    def yield_image_file_sequence():
        for img in menu[i_position % num_items].get_image_files():
            yield img
        for i in range(1, 12):
            # if (i_position - i) >= min_item:
            #     for img in menu[i_position - i].get_image_files():
            #         yield img
            # if (i_position + i) <= max_item:
            #     for img in menu[i_position + i].get_image_files():
            #         yield img
            for img in menu[(i_position - i) % num_items].get_image_files():
                yield img
            for img in menu[(i_position + i) % num_items].get_image_files():
                yield img

    item_data = []
    if "ITEMS" in what:
        TextureManager.get().load_images(list(yield_image_file_sequence()))
        # TextureManager.get().load_textures(1)

        center_item = None
        light1_position = [-1.5, 0.2, -2.3, 1.0]
        light2_position = [-0.5, 2.5, -2.3, 1.0]
        # glDisable(GL_LIGHT2)

        if LIGHTING:
            gl.glEnable(gl.GL_LIGHTING)
            gl.glLightfv(gl.GL_LIGHT1, gl.GL_POSITION, light1_position)
            gl.glLightfv(gl.GL_LIGHT2, gl.GL_POSITION, light2_position)
            gl.glMaterialfv(gl.GL_FRONT, gl.GL_DIFFUSE, (1.0, 1.0, 1.0, 1.0))
        gl.glColor3f(1.0, 1.0, 1.0)

        # gloss_list = []
        # reflections = []
        # glColor3f(1.0, 1.0, 1.0)

        fs_emu_texturing(True)
        fs_emu_blending(False)

        for item_index in list(yield_render_item_sequence()):
            if num_items < MIN_WRAPAROUND_ITEMS:
                if not (
                    0
                    <= item_index - (i_position // num_items * num_items)
                    < num_items
                ):
                    continue

            rel_index = item_index - position  # selected

            distance = abs(rel_index)
            # strength = 1.0

            gl.glPushMatrix()

            pos, rotation = get_item_position(rel_index)

            gl.glTranslate(*pos)
            scale = 1.0
            area = 0.9

            item = menu[item_index % num_items]
            height = math.sqrt(area / item.ratio)
            width = item.ratio * height

            if rel_index <= 0.0:
                spec = 0.8
            elif rel_index > 1.0:
                spec = 0.05
            else:
                spec = 0.8 - rel_index / 1.0 * 0.75
            if LIGHTING:
                gl.glMaterialfv(
                    gl.GL_FRONT, gl.GL_SPECULAR, (spec, spec, spec, 1.0)
                )

            gl.glRotate(rotation, 0.0, 1.0, 0.0)

            item.render(width, height)

            if not LIGHTING:
                if distance < 1.0:
                    gloss_alpha = 0.25 + distance * 0.25
                else:
                    gloss_alpha = 0.5
                render_item_gloss(gloss_alpha, ratio=item.ratio, area=area)
            gl.glPopMatrix()
            item_data.append(
                (
                    "ITEM",
                    menu,
                    item_index,
                    pos,
                    rotation,
                    scale,
                    item.ratio,
                    1.0,
                    area,
                )
            )

            # reflections.append((item, pos, rotation,
            # width, height, spec * 0.3))

        if not center_item:
            State.get().center_item = None

        if LIGHTING:
            light1_position[1] = -light1_position[1]
            light2_position[1] = -light2_position[1]
            gl.glLightfv(gl.GL_LIGHT1, gl.GL_POSITION, light1_position)
            gl.glLightfv(gl.GL_LIGHT2, gl.GL_POSITION, light2_position)
            # glLightfv(GL_LIGHT1, GL_POSITION, (-5.0, -1.5, 1.2, 1.0))
            # glLightfv(GL_LIGHT2, GL_POSITION, (-0.5, -5.0, 1.2, 1.0))
            gl.glMaterialfv(gl.GL_FRONT, gl.GL_DIFFUSE, (0.3, 0.3, 0.3, 0.0))

        # glColor3f(0.3, 0.3, 0.3)
        # for item, pos, rotation, width, height, spec in reflections:
        #     #if LIGHTING:
        #     #    glMaterialfv(GL_FRONT, GL_SPECULAR, (spec, spec, spec, 1.0))
        #     #else:
        #     #    glColor4f(0.33, 0.33, 0.33, 0.33)
        #     #    fs_emu_blending(True)
        #     glPushMatrix()
        #     glTranslate(*pos)
        #     glRotate(rotation, 0.0, 1.0, 0.0)
        #     item.render(width, height, reflection=True)
        #     glPopMatrix()

        if LIGHTING:
            gl.glDisable(gl.GL_LIGHTING)

    return item_data


def render_item_gloss(alpha, ratio=1.0, brightness=1.0, area=None):
    if State.get().max_ratio:
        ratio = min(State.get().max_ratio, ratio)
    if State.get().force_portrait and ratio > 1.0:
        ratio = 1.0 / ratio
    ba = brightness * alpha
    gl.glColor4f(ba, ba, ba, alpha)
    fs_emu_blending(True)
    fs_emu_texturing(True)
    creating_list = DisplayLists.call_or_create("gloss", ratio)
    if not creating_list:
        return
    gl.glPolygonOffset(0.0, -4.0)
    gl.glEnable(gl.GL_POLYGON_OFFSET_FILL)
    gl.glDepthMask(False)
    # glBindTexture(GL_TEXTURE_2D, Texture.gloss)
    texture = Texture.gloss
    texture.bind()
    area = area or 0.8
    height = math.sqrt(area / ratio)
    width = ratio * height
    gl.glTranslatef(-width / 2.0, -0.5, 0.0)
    gl.glBegin(gl.GL_QUADS)
    gl.glTexCoord(0.0, 1.0)
    gl.glVertex3f(0.0, 0.0, 0.0)
    gl.glTexCoord(1.0, 1.0)
    gl.glVertex3f(width, 0.0, 0.0)
    gl.glTexCoord(1.0, 0.0)
    gl.glVertex3f(width, height, 0.0)
    gl.glTexCoord(0.0, 0.0)
    gl.glVertex3f(0.0, height, 0.0)
    gl.glEnd()
    gl.glPolygonOffset(0.0, 0.0)
    gl.glDisable(gl.GL_POLYGON_OFFSET_FILL)
    gl.glDepthMask(True)
    # fs_emu_blending(False)
    if creating_list:
        gl.glEndList()


def render_item_shadows(itemdata, front=False, back=True):
    Render.get().standard_perspective()
    # glEnable(GL_STENCIL_TEST)
    # glStencilFunc(GL_EQUAL, 1, 1)
    if front:
        gl.glPolygonOffset(0.0, -4.0)
        gl.glEnable(gl.GL_POLYGON_OFFSET_FILL)
    elif back:
        gl.glPolygonOffset(0.0, 4.0)
        gl.glEnable(gl.GL_POLYGON_OFFSET_FILL)
    gl.glDepthMask(False)
    fs_emu_blending(True)
    fs_emu_texturing(True)
    gl.glBindTexture(gl.GL_TEXTURE_2D, Texture.shadow2)

    for (
        type_,
        menu,
        item_index,
        pos,
        rotation,
        scale,
        ratio,
        brightness,
        area,
    ) in reversed(itemdata):
        gl.glPushMatrix()
        gl.glTranslate(*pos)
        gl.glRotate(rotation, 0.0, 1.0, 0.0)
        gl.glScalef(scale, scale, 1.0)
        render_item_shadow(ratio=ratio, brightness=brightness, area=area)
        gl.glScalef(1.0, -1.0, 1.0)
        gl.glTranslatef(0.0, 0.5, 0.0)
        # render_item_shadow(ratio=ratio,
        # brightness=brightness * 0.33, area=area)
        render_item_shadow(ratio=ratio, brightness=brightness, area=area)
        gl.glPopMatrix()

    gl.glPolygonOffset(0.0, 0.0)
    gl.glDisable(gl.GL_POLYGON_OFFSET_FILL)
    gl.glDepthMask(True)
    # fs_emu_blending(False)
    # glDisable(GL_STENCIL_TEST)


def render_item_shadow(ratio=1.0, brightness=1.0, area=None):
    if State.get().max_ratio:
        ratio = min(State.get().max_ratio, ratio)
    if State.get().force_portrait and ratio > 1.0:
        ratio = 1.0 / ratio
    gl.glColor4f(brightness, brightness, brightness, brightness)
    creating_list = DisplayLists.call_or_create("shadow", ratio, area)
    if not creating_list:
        return
    gl.glTranslatef(0.0, -0.5, 0.0)
    area = area or 0.8
    height = math.sqrt(area / ratio)
    width = ratio * height
    wh = width / 2
    t1 = 120 / 512
    t2 = (512 - 120) / 512
    gl.glBegin(gl.GL_QUADS)
    # left
    gl.glTexCoord(0.0, t2)
    gl.glVertex3f(-wh - 0.1, 0.02, 0.0)
    gl.glTexCoord(t1, t2)
    gl.glVertex3f(-wh + 0.02, 0.02, 0.0)
    gl.glTexCoord(t1, t1)
    gl.glVertex3f(-wh + 0.02, height - 0.02, 0.0)
    gl.glTexCoord(0.0, t1)
    gl.glVertex3f(-wh - 0.1, height - 0.02, 0.0)
    # top left
    gl.glTexCoord(0.0, t1)
    gl.glVertex3f(-wh - 0.1, height - 0.02, 0.0)
    gl.glTexCoord(t1, t1)
    gl.glVertex3f(-wh + 0.02, height - 0.02, 0.0)
    gl.glTexCoord(t1, 0.0)
    gl.glVertex3f(-wh + 0.02, height + 0.1, 0.0)
    gl.glTexCoord(0.0, 0.0)
    gl.glVertex3f(-wh - 0.1, height + 0.1, 0.0)
    # top
    gl.glTexCoord(t1, t1)
    gl.glVertex3f(-wh + 0.02, height - 0.02, 0.0)
    gl.glTexCoord(t2, t1)
    gl.glVertex3f(wh - 0.02, height - 0.02, 0.0)
    gl.glTexCoord(t2, 0.0)
    gl.glVertex3f(wh - 0.02, height + 0.1, 0.0)
    gl.glTexCoord(t1, 0.0)
    gl.glVertex3f(-wh + 0.02, height + 0.1, 0.0)
    # top right
    gl.glTexCoord(t2, t1)
    gl.glVertex3f(wh - 0.02, height - 0.02, 0.0)
    gl.glTexCoord(1.0, t1)
    gl.glVertex3f(wh + 0.1, height - 0.02, 0.0)
    gl.glTexCoord(1.0, 0.0)
    gl.glVertex3f(wh + 0.1, height + 0.1, 0.0)
    gl.glTexCoord(t2, 0.0)
    gl.glVertex3f(wh - 0.02, height + 0.1, 0.0)
    # right
    gl.glTexCoord(t2, t2)
    gl.glVertex3f(wh - 0.02, 0.02, 0.0)
    gl.glTexCoord(1.0, t2)
    gl.glVertex3f(wh + 0.1, 0.02, 0.0)
    gl.glTexCoord(1.0, t1)
    gl.glVertex3f(wh + 0.1, height - 0.02, 0.0)
    gl.glTexCoord(t2, t1)
    gl.glVertex3f(wh - 0.02, height - 0.02, 0.0)
    # bottom right
    gl.glTexCoord(t2, 1.0)
    gl.glVertex3f(wh - 0.02, -0.1, 0.0)
    gl.glTexCoord(1.0, 1.0)
    gl.glVertex3f(wh + 0.1, -0.1, 0.0)
    gl.glTexCoord(1.0, t2)
    gl.glVertex3f(wh + 0.1, 0.02, 0.0)
    gl.glTexCoord(t2, t2)
    gl.glVertex3f(wh - 0.02, 0.02, 0.0)
    # bottom
    gl.glTexCoord(t1, 1.0)
    gl.glVertex3f(-wh + 0.02, -0.1, 0.0)
    gl.glTexCoord(t2, 1.0)
    gl.glVertex3f(wh - 0.02, -0.1, 0.0)
    gl.glTexCoord(t2, t2)
    gl.glVertex3f(wh - 0.02, 0.02, 0.0)
    gl.glTexCoord(t1, t2)
    gl.glVertex3f(-wh + 0.02, 0.02, 0.0)
    # bottom left
    gl.glTexCoord(0.0, 1.0)
    gl.glVertex3f(-wh - 0.1, -0.1, 0.0)
    gl.glTexCoord(t1, 1.0)
    gl.glVertex3f(-wh + 0.02, -0.1, 0.0)
    gl.glTexCoord(t1, t2)
    gl.glVertex3f(-wh + 0.02, 0.02, 0.0)
    gl.glTexCoord(0.0, t2)
    gl.glVertex3f(-wh - 0.1, 0.02, 0.0)
    gl.glEnd()
    # end list
    if creating_list:
        gl.glEndList()


class ItemMenu(Menu):
    def __init__(self, title=""):
        Menu.__init__(self, title)

    def render(self):
        render_bottom_bar()
        result = render_menu(self)
        # FIXME: NO NEED TO RENDER THE BOTTOM PART OF THE WALL
        # -flag to render_wall?

        # from .gamemenu import render_wall
        # render_wall()
        return result

    def render_transparent(self, data):
        render_item_shadows(data)
        render_bottom_bar_transparent()
        pass



from fsgs.drivers.messdriver import MessDriver
from fsgs.option import Option
from fsgs.spectrum import (
    ZXS_48_ROM,
    ZXS_128_0_ROM,
    ZXS_128_1_ROM,
    ZXS_PLUS3_0_ROM,
    ZXS_PLUS3_1_ROM,
    ZXS_PLUS3_2_ROM,
    ZXS_PLUS3_3_ROM,
)

MESS_SPECTRUM = "spectrum"  # ZX Spectrum
MESS_SPEC128 = "spec128"  # ZX Spectrum 128
MESS_SPECPLS3 = "specpls3"  # ZX Spectrum +3
MESS_sp3e8bit = "sp3e8bit"  # ZX Spectrum +3e 8bit IDE
# noinspection SpellCheckingInspection
MESS_sp3eata = "sp3eata"  # ZX Spectrum +3e 8bit ZXATASP
# noinspection SpellCheckingInspection
MESS_sp3ezcf = "sp3ezcf"  # ZX Spectrum +3e 8bit ZXCF
MESS_spec80k = "spec80k"  # ZX Spectrum 80K
MESS_specide = "specide"  # ZX Spectrum IDE
MESS_specpl2a = "specpl2a"  # ZX Spectrum +2a
MESS_specpl3e = "specpl3e"  # ZX Spectrum +3e
MESS_specpls2 = "specpls2"  # ZX Spectrum +2


class MessSpectrumDriver(MessDriver):
    JOYSTICK = {
        "type": "joystick",
        "description": "Joystick",
        "mapping_name": "zx-spectrum-joystick",
    }

    PORTS = [{"description": "Joystick Port", "types": [JOYSTICK]}]

    def mess_configure(self):
        # FIXME: hack
        # self.mess_configure_floppies(["cassette"])
        # self.emulator.args.extend(["-ui_active"])

        if self.config[Option.SNAPSHOT_FILE]:
            # snapshot loads automatically, no need to do anything special
            self.add_arg(
                "-{0}".format("snapshot"),
                self.get_game_file(Option.SNAPSHOT_FILE),
            )
        elif self.config[Option.FLOPPY_DRIVE_0]:
            # pressing enter to select "loader" in the menu
            self.inject_fake_input_string(160, "{0}\n".format(""))
            self.add_arg(
                "-{0}".format("floppydisk1"),
                self.get_game_file(Option.FLOPPY_DRIVE_0),
            )

        elif self.config[Option.TAPE_DRIVE_0]:
            self.emulator.args.extend(
                ["-cassette", self.get_game_file(Option.TAPE_DRIVE_0)]
            )
            if self.config[Option.ZXS_MODEL] == "spectrum128":
                self.inject_fake_input_string_list(
                    160,
                    [
                        "1040",
                        "0040",
                        "0000",  # Return
                        "0000",
                        "0000",
                        "0000",
                        "1059",
                        "0059",
                        "0000",  # F2 (Load tape)
                    ],
                )
            else:
                self.inject_fake_input_string_list(
                    160,
                    [
                        # Inject LOAD ""
                        "1013",
                        "0013",
                        "0000",  # J (LOAD)
                        "1229",
                        "0000",  # Right Shift (press)
                        "1019",
                        "0019",
                        "0000",  # P (with symbol shift = ")
                        "0229",
                        "0000",  # Right shift (release)
                        "1229",
                        "0000",  # Right shift (press)
                        "1019",
                        "0019",
                        "0000",  # P (with symbol shift = ")
                        "0299",
                        "0000",  # Right shift (release)
                        "1040",
                        "0040",
                        "0000",  # Return
                        # "0000", "0000", "0000",
                        # "1071", "0071", "0000",  # Scroll Lock
                        "0000",
                        "0000",
                        "0000",
                        "1059",
                        "0059",
                        "0000",  # F2 (Load tape)
                    ],
                )
            # self.args.extend(
            #     ["-autoboot_command", 'j" "\\n'])

    def mess_full_keyboard(self):
        return False

    def mess_input_mapping(self, port):
        return {
            "UP": 'tag=":KEMPSTON" type="KEYBOARD" mask="8" defvalue="0"',
            "DOWN": 'tag=":KEMPSTON" type="KEYBOARD" mask="4" defvalue="0"',
            "LEFT": 'tag=":KEMPSTON" type="KEYBOARD" mask="2" defvalue="0"',
            "RIGHT": 'tag=":KEMPSTON" type="KEYBOARD" mask="1" defvalue="0"',
            "1": 'tag=":KEMPSTON" type="KEYBOARD" mask="16" defvalue="0"',
        }

    def get_game_refresh_rate(self):
        # Refresh rate values retrieved from MESS
        if self.config[Option.ZXS_MODEL] == "spectrum128":
            return 50.021084
        else:
            return 50.080128

    def mess_romset(self):
        if self.config[Option.ZXS_MODEL] == "spectrum128":
            return MESS_SPEC128, MESS_SPEC128_ROMS
        if self.config[Option.ZXS_MODEL] == "spectrum+3":
            return MESS_SPECPLS3, MESS_SPECPLS3_ROMS
        else:
            return MESS_SPECTRUM, MESS_SPECTRUM_ROMS


# noinspection SpellCheckingInspection
MESS_SPECTRUM_ROMS = [
    (ZXS_48_ROM.sha1, "spectrum.rom"),
    # ("f9d23f25640c51bcaa63e21ed5dd66bb2d5f63d4", "1986es.rom"),
    # 9e535e2e24231ccb65e33d107f6d0ceb23e99477", "48e.rom"),
    # ("e62a431b0938af414b7ab8b1349a18b3c4407f70", "48turbo.rom"),
    # ("ab3c36daad4325c1d3b907b6dc9a14af483d14ec", "bsrom118.rom"),
    # ("2ee2dbe6ab96b60d7af1d6cb763b299374c21776", "bsrom140.rom"),
    # ("795c20324311dd5a56300e6e4ec49b0a694ac0b3", "deutsch.rom"),
    # ("51165cde68e218512d3145467074bc7e786bf307", "groot.rom"),
    # ("a701c3d4b698f7d2be537dc6f79e06e4dbc95929", "gw03.rom"),
    # ("2a9745ba3b369a84c4913c98ede66ec87cb8aec1", "hdt-iso.rom"),
    # ("dee814271c4d51de257d88128acdb324fb1d1d0d", "imc.rom"),
    # ("5752e6f789769475711b91e0a75911fa5232c767", "iso8bm.rom"),
    # ("04adbdb1380d6ccd4ab26ddd61b9ccbba462a60f", "isomoje.rom"),
    # ("d7f02ed66455f1c08ac0c864c7038a92a88ba94a", "jgh.rom"),
    # ("c103e89ef58e6ade0c01cea0247b332623bd9a30", "plus4.rom"),
    # ("0853e25857d51dd41b20a6dbc8e80f028c5befaa", "psycho.rom"),
    # ("c58ff44a28db47400f09ed362ca0527591218136", "sc01.rom"),
    # ("84ea64af06adaf05e68abe1d69454b4fc6888505", "turbo2_3.rom"),
    # ("21ad93ffe41a4458704c866cca2754f066f6a560", "turbo4_4.rom"),
]

# noinspection SpellCheckingInspection
MESS_SPEC128_ROMS = [
    (ZXS_128_0_ROM.sha1, "zx128_0.rom"),
    (ZXS_128_1_ROM.sha1, "zx128_1.rom"),
    # Spanish
    # ("968937b1c750f0ef6205f01c6db4148da4cca4e3", "zx128s0.rom"),
    # ("bea3f397cc705eafee995ea629f4a82550562f90", "zx128s1.rom"),
]

# noinspection SpellCheckingInspection
MESS_SPECPLS3_ROMS = [
    (ZXS_PLUS3_0_ROM.sha1, "pl3-0.rom"),
    (ZXS_PLUS3_1_ROM.sha1, "pl3-1.rom"),
    (ZXS_PLUS3_2_ROM.sha1, "pl3-2.rom"),
    (ZXS_PLUS3_3_ROM.sha1, "pl3-3.rom"),
    # ("4e5d114b72d464cefdde0566457f52a3c0c1cae2", "p3_01_4m.rom"),
    # ("4e5d114b72d464cefdde0566457f52a3c0c1cae2", "p3_01_cm.rom"),
    # ("752cdd6a083ab9910348995e483541d60bb6372b", "p3_23_4m.rom"),
    # ("d062765ceb1f3cd2c94ea51cb737cac7ad6151b4", "p3_23_cm.rom"),
    # ("500c0945760abeefcbd08bc22c0d07b14b336cf0", "plus341.rom"),
    # ("e9b0a60a1a8def511d59090b945d175bdc646346", "plus3sp0.rom"),
    # ("4e48f196427596c7990c175d135c15a039c274a4", "plus3sp1.rom"),
    # ("09fc005625589ef5992515957ce7a3167dec24b2", "plus3sp2.rom"),
    # ("ec8f644a81e2e9bcb58ace974103ea960361bad2", "plus3sp3.rom"),
]



import os

from fsbc.system import windows

# import pygame
from fsbc.user import get_documents_dir, get_home_dir
from fsbc.util import memoize


@memoize
def get_fullscreen_size():
    # pygame.display.init()
    # display_info = pygame.display.Info()
    # size = (display_info.current_w, display_info.current_h)
    # print("fullscreen size:", size)
    # return size
    pass


class Settings(object):
    cache_dir_path = None
    games_dir_path = None
    profile_dir_path = None
    # use_fullscreen = False

    if windows:
        # fullscreen_menu = False
        window_decorations = False
    else:
        # fullscreen_menu = True
        window_decorations = True

    fullscreen_menu = True
    fullscreen_game = True
    # windowed_size = (640, 480)
    windowed_size = None  # Maximized
    fullscreen_size = get_fullscreen_size()
    fullscreen_pos = 0, 0

    @classmethod
    def get_games_path(cls):
        path = []
        if cls.games_dir_path:
            path.extend(cls.games_dir_path)
        elif windows:
            path.append(
                os.path.join(get_documents_dir(), "Games (Ku Game System)")
            )
        else:
            path.append(os.path.join(get_home_dir(), "Games"))
        return path

    @classmethod
    def get_resources_dir(cls):
        path = os.path.join(cls.get_games_path()[0], "Resources")
        return path

    @classmethod
    def get_config_dir(cls):
        pass

    @classmethod
    def get_profile_dir(cls):
        if cls.profile_dir_path:
            return cls.profile_dir_path
        elif windows:
            return os.path.join(
                get_documents_dir(), "Gamer Profile (Ku Game System)"
            )
        else:
            return os.path.join(
                get_home_dir(), "Gamer Profile (Ku Game System)"
            )



class IRCBroadcaster:
    irc_listeners = []

    @classmethod
    def add_listener(cls, listener):
        cls.irc_listeners.append(listener)

    @classmethod
    def remove_listener(cls, listener):
        cls.irc_listeners.remove(listener)

    @classmethod
    def broadcast(cls, key, value):
        for listener in cls.irc_listeners:
            listener.on_irc(key, value)



import fsui as fsui
from .WidgetTestCase import WidgetTestCase


class TestButton(WidgetTestCase):
    def test_button(self):
        self.setup_dialog()
        button = fsui.Button(self.dialog)
        button.disable()
        button.enable()
        button.enable(False)
        button.enable(True)
        self.run_dialog()

    def test_button_with_label(self):
        self.setup_dialog()
        fsui.Button(self.dialog, "With Label")
        self.run_dialog()



import atexit
import json
import subprocess
import traceback
from queue import Empty, Queue
from threading import Thread

from fsgs.amiga.fsuaedevicehelper import FSUAEDeviceHelper


class EventHelper(Thread):
    def __init__(self):
        Thread.__init__(self, name="EventHelperThread")
        self.setDaemon(True)

        self.devices = []
        self.joystick_devices = []
        self.keyboard_devices = []
        self.joystick_like_devices = []
        self.initialized = False

        self.process = None
        self.events = Queue()
        self.stopping = False

    def init(self):
        if self.initialized:
            return
        self.init_device_helper()
        self.initialized = True

    def deinit(self):
        self.stopping = True

    def is_add_event(self, event):
        return event["type"] in [
            "joy-device-added",
            "mouse-device-added",
            "keyboard-device-added",
        ]

    def is_remove_event(self, event):
        return event["type"] in [
            "joy-device-removed",
            "mouse-device-removed",
            "keyboard-device-removed",
        ]

    def get_next_event(self):
        try:
            event = self.events.get(False)
            return event
        except Empty:
            return None

    def run(self):
        try:
            self._run()
        except Exception:
            traceback.print_exc()
        try:
            self.process.kill()
        except Exception:
            traceback.print_exc()
        self.process = None

    def _run(self):
        print("[INPUT] EventHelper thread started")
        while not self.stopping:
            line = self.process.stdout.readline()
            line = line.decode("UTF-8", "replace")
            if not line:
                print("no line returned from device helper")
                break
            if line.startswith("#"):
                continue
            try:
                event = json.loads(line)
            except Exception:
                print("Problem loading JSON from line:", line)
                continue
            # print(event)
            self.events.put(event)
        print("[INPUT] EventHelper thread stopping")
        self.process.kill()
        atexit.unregister(self.kill_device_helper_on_exit)

    def init_device_helper(self):
        print("EventHelper.init_device_helper")
        try:
            self.process = FSUAEDeviceHelper.start_with_args(
                ["--events"], stdout=subprocess.PIPE
            )
        except Exception:
            print("exception while listing joysticks and devices")
            traceback.print_exc()
            return

        atexit.register(self.kill_device_helper_on_exit)
        self.start()

    def kill_device_helper_on_exit(self):
        print("[INPUT] Killing device helper process (on exit)")
        self.process.kill()



import functools
import os
import sys
import unicodedata

import fsboot
from fsbc.system import macosx, windows


class Paths(object):
    @staticmethod
    def str(path):
        return path.encode(sys.getfilesystemencoding())

    @staticmethod
    def encode(path):
        return path.encode(sys.getfilesystemencoding())

    @staticmethod
    def unicode(path):
        if isinstance(path, str):
            return path
        return path.decode(sys.getfilesystemencoding())

    @classmethod
    def join(cls, a, b):
        # if not a:
        #     return b
        # if a[-1] == "/" or a[-1] == "\\":
        #     return a + b
        # return a + "/" + b
        return os.path.join(a, b).replace("\\", "/")

    @classmethod
    def expand_path(cls, path, default_dir=None):
        if path and path[0] == "$":
            cmp_path = path.upper().replace("\\", "/")
            if cmp_path.startswith("$BASE/"):
                return cls.join(cls.get_base_dir(), path[6:])
            if cmp_path.startswith("$CONFIG/"):
                # FIXME: dependency loop, have FS-UAE Launcher register
                # this prefix with this class instead
                from launcher.launcher_settings import LauncherSettings

                config_path = LauncherSettings.get("config_path")
                if config_path:
                    return cls.join(os.path.dirname(config_path), path[8:])
            if cmp_path.startswith("$HOME/"):
                return cls.join(cls.get_home_dir(), path[6:])
            # FIXME: $base_dir is deprecated
            if cmp_path.startswith("$BASE_DIR/"):
                return cls.join(cls.get_base_dir(), path[10:])
        elif not os.path.isabs(path) and default_dir is not None:
            return os.path.join(default_dir, path)
        return path

    @classmethod
    def contract_path(cls, path, default_dir=None, force_real_case=True):
        if path.rfind(":") > 1:
            # Checking against > index 1 to allow for Windows absolute paths
            # with drive letter and colon. If colon is later, we assume this
            # is an URI, and not a path, so we do not do anything with it
            return path
        if force_real_case:
            print("before", path)
            path = cls.get_real_case(path)
            print("after", path)
        # dir, file = os.path.split(path)
        # norm_dir = dir + "/"
        if default_dir is not None:
            default_dir += "/"
            if path.startswith(default_dir):
                return path[len(default_dir) :]
        base_dir = cls.get_base_dir(slash=True)
        if path.startswith(base_dir):
            path = path.replace(base_dir, "$BASE/")
        home_dir = cls.get_home_dir(slash=True)
        if path.startswith(home_dir):
            path = path.replace(home_dir, "$HOME/")
        return path

    @classmethod
    def get_base_dir(cls, slash=False):
        path = fsboot.base_dir()
        path = cls.get_real_case(path)
        if slash:
            path += "/"
        return path

    @classmethod
    @functools.lru_cache()
    def get_home_dir(cls, slash=False):
        path = fsboot.home_dir()
        path = cls.get_real_case(path)
        if slash:
            path += "/"
        return path

    @classmethod
    def get_real_case(cls, path):
        """Check the case for the (case insensitive) path. Used to make the
        database portable across sensitive/insensitive file systems."""

        # get_real_case will fail on Linux if you have "conflicting" paths
        # (differing only by case), unless we check that the specified path
        # is already correct. The reason for not simply returning the path
        # as-is on Linux, is to allow this function can find files in
        # directories (portable version) when the directory is specified with
        # incorrect case.
        if not windows and not macosx:
            if os.path.exists(path):
                return path

        parts = []
        drive, p = os.path.splitdrive(path)
        removed_separator = ""
        if path.startswith("/"):
            drive = "/"
        elif drive:
            # on Windows, add / to make drive a valid path
            drive += "/"
        if len(p) > 1 and (p.endswith("/") or (windows and p.endswith("\\"))):
            removed_separator = p[-1]
            p = p[:-1]

        last = ""
        while p != last:
            name = os.path.basename(p)
            if not name:
                break
            parts.append(name)
            last = p
            p = os.path.dirname(p)
        parts.reverse()
        result = [drive]
        result.extend(parts)

        combined = drive
        combined = combined.upper()
        k = 1
        for part in parts:
            part_compare = part
            part_compare = part_compare.lower()
            if macosx:
                part_compare = unicodedata.normalize("NFC", part_compare)
            # print("part is", part)
            if os.path.isdir(combined):
                # print("checking case of", combined + "/" + part)
                for name in os.listdir(combined):
                    # if part == "Førde":
                    #     print(os.listdir(combined))
                    name_compare = name
                    name_compare = name_compare.lower()
                    if macosx:
                        name_compare = unicodedata.normalize(
                            "NFC", name_compare
                        )
                    if name_compare == part_compare:
                        # print("found case =", name)
                        if not combined.endswith("/"):
                            combined += "/"
                        combined += name
                        result[k] = name
                        break
                else:
                    raise Exception("could not find case for path " + path)
            k += 1

        # FIXME: could be an idea to always normalize to NFC on OS X too,
        # to make the database even more portable

        # normalizing slashes to forward slash to make the database more
        # portable
        result_path = os.path.join(*result).replace("\\", "/")
        result_path += removed_separator
        return result_path



import os
import shutil
import warnings
from configparser import ConfigParser

from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.GameChangeHandler import GameChangeHandler
from fsgs.util.gamenameutil import GameNameUtil

"""
TODO: Snapshots support. Create a Snapshots/TimeTDate/ directory with a 
copy of the state dir!
"""


class SaveHandler(object):
    def __init__(
        self,
        fsgc,
        name="",
        platform="Unknown",
        uuid=None,
        options=None,
        emulator="",
    ):
        self.fsgc = fsgc
        # FIXME: SaveHandler(fsgc, options=options) is new usage?
        self._options = options
        self._emulator_specific = False
        self._emulator_name = emulator

        self.uuid = uuid
        self.change_handlers = []

        self.config_name = name
        if "(" in name:
            parts = name.split("(", 1)
            self.name, self.variant = parts
            self.name = self.name.strip()
            self.variant = self.variant.strip()
            if self.variant.endswith(")"):
                self.variant = self.variant[:-1]
            self.variant = self.variant.replace(") (", ", ")
            self.variant = self.variant.replace(")(", ", ")
        else:
            self.name = name
            self.variant = ""
        self.platform = platform

    def set_save_data_is_emulator_specific(self, emulator_specific=True):
        self._emulator_specific = emulator_specific

    def get(self, name):
        if self._options is not None:
            return self._options.get(name, "")
        value = self.fsgc.config.get(name, "")
        if not value:
            value = self.fsgc.settings.get(name, "")
        return value

    def register_changes(self, original_dir, changes_dir):
        self.change_handlers.append(
            (GameChangeHandler(original_dir), changes_dir)
        )

    def prepare(self):
        self.create_save_ini()
        for change_handler, changes_dir in self.change_handlers:
            change_handler.init(changes_dir)

    def create_save_ini(self):
        save_dir = self.base_save_dir()
        save_ini_path = os.path.join(save_dir, "Save.ini")

        cp = ConfigParser()
        if os.path.exists(save_ini_path):
            with open(save_ini_path, "r", encoding="UTF-8") as f:
                cp.read(f)

        if not cp.has_section("fsgs-save"):
            cp.add_section("fsgs-save")
        cp.set("fsgs-save", "version", "2")
        cp.set("fsgs-save", "platform", self.get("platform"))
        cp.set("fsgs-save", "game_uuid", self.get("game_uuid"))
        cp.set("fsgs-save", "game_name", self.get("game_name"))
        cp.set("fsgs-save", "variant_uuid", self.get("variant_uuid"))
        cp.set("fsgs-save", "variant_name", self.get("variant_name"))

        with open(save_ini_path + ".partial", "w", encoding="UTF-8") as f:
            cp.write(f)
        os.replace(save_ini_path + ".partial", save_ini_path)

    def base_save_dir(self):
        save_dir = self.save_dir_path()
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        return save_dir

    def save_dir(self):
        return self.base_save_dir()

    def emulator_state_dir(self, name=""):
        emulator = name or self._emulator_name
        assert emulator
        save_state_directory = os.path.join(self.base_save_dir(), emulator)
        if not os.path.exists(save_state_directory):
            os.makedirs(save_state_directory)
        return save_state_directory

    def emulator_save_dir(self, name=""):
        emulator = name or self._emulator_name
        if self._emulator_specific:
            return self.emulator_state_dir(emulator)
        else:
            return self.base_save_dir()

    def uuid_save_dir_path(self):
        saves_dir = FSGSDirectories.saves_dir()
        # FIXME: Correct?
        variant_uuid = self.get("variant_uuid")
        if variant_uuid:
            save_dir = os.path.join(
                saves_dir, "UUID", variant_uuid[:2], variant_uuid
            )
            return save_dir
        return None

    def save_dir_path(self):
        save_dir = self.uuid_save_dir_path()
        if save_dir:
            return save_dir
        # saves_dir = FSGSDirectories.saves_dir()
        assert False, "save_dir_path not implemented for this case"

    def finish(self):
        for change_handler, changes_dir in self.change_handlers:
            change_handler.update(changes_dir)
        self.cleanup()

    def cleanup(self):
        save_dir = self.save_dir_path()
        print("[SAVES] Cleanup", save_dir)
        if not os.path.exists(save_dir):
            print("[SAVES] Save dir does not exist, no need to clean up")
            return
        erase = True
        for dir_path, dir_name, file_names in os.walk(save_dir):
            for file_name in file_names:
                if file_name != "Save.ini":
                    erase = False
        if not erase:
            print("[SAVES] Save data found, will not erase save dir")
            for item in os.listdir(save_dir):
                sub_dir = os.path.join(save_dir, item)
                if os.path.isdir(sub_dir):
                    self.cleanup_sub_dir(sub_dir)
            return
        print("[SAVES] No save data found, erasing save dir!")
        shutil.rmtree(save_dir)
        # FIXME: Remove UUID/XX dir if empty?

    def cleanup_sub_dir(self, sub_dir):
        """Check if there are any files in this directory sub tree. If not,
        then remove the directory."""
        for dir_path, dir_name, file_names in os.walk(sub_dir):
            for _ in file_names:
                # At least one file was found
                return
        print("[SAVES] Removing empty dir:", sub_dir)
        shutil.rmtree(sub_dir)

    # -------------------------------------------------------------- legacy ---

    @property
    def fsgs(self):
        warnings.warn("Use .fsgc instead", DeprecationWarning)
        return self.fsgc

    def get_name(self):
        # warnings.warn("Use .name instead", DeprecationWarning)
        return self.name

    def get_variant(self):
        # warnings.warn("Use .variant instead", DeprecationWarning)
        return self.variant

    def _get_state_dir(self):
        config_name = self.config_name
        if not config_name:
            config_name = "Default"

        # Use a temporary state dir, for now, to avoid problems with
        # floppy overlays etc interfering with net play.
        if self.fsgc.netplay.enabled:
            # It is possible to manually specify the state dir.
            config_name = self.fsgc.config.get("__netplay_state_dir_name")
            if not config_name:
                # This is the default behavior, create a clean state
                # dir for the net play session.
                netplay_game = self.fsgc.config.get("__netplay_game")
                if netplay_game:
                    config_name = "Net Play ({0})".format(netplay_game)

        # Convert the config name to a name which can be represented on
        # the file system (really all/most filesystems).
        config_name = GameNameUtil.create_fs_name(config_name)

        letter = self.get_letter(config_name)
        if not letter:
            config_name = "Default"
            letter = self.get_letter(config_name)
        # We use an existing state dir in a "letter" dir if it already
        # exists (legacy support).
        path = os.path.join(
            FSGSDirectories.get_save_states_dir(), letter, config_name
        )
        if os.path.exists(path):
            return path
        # If not, we use a direct sub-folder of save states dir.
        path = os.path.join(FSGSDirectories.get_save_states_dir(), config_name)
        return path

    def get_state_dir(self):
        state_dir = self._get_state_dir()
        if not os.path.exists(state_dir):
            os.makedirs(state_dir)
        return state_dir

    @staticmethod
    def get_letter(name):
        letter_name = name.upper()
        if letter_name.startswith("THE "):
            letter_name = letter_name[4:]
        if letter_name.startswith("A "):
            letter_name = letter_name[2:]
        for i in range(len(letter_name)):
            letter = letter_name[i]
            if letter in "01234567890":
                letter = "0"
                break
            if letter in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
                break
        else:
            return None
        return letter
